from dataclasses import dataclass
from enum import Enum
from typing import Optional


class DeploymentState(Enum):
    """The state of a deployment request. One of:

    - PENDING: the request has been created and is pending approval
    - NOTIFIED: the local admin has been notified of the pending request
    - REJECTED: the request has been rejected
    - UNKNOWN: the request state is unknown

    Once it is accepted, the request is removed from the pending database.
    """

    PENDING = 0
    NOTIFIED = 1
    REJECTED = 2
    UNKNOWN = 3

    @staticmethod
    def from_string(state: str):
        """
        Convert a string to a DeploymentState.
        Args:
            state (str): the string to convert
        Returns:
            DeploymentState: the DeploymentState or DeploymentState.UNKNOWN if the string is not a valid DeploymentState
        """
        try:
            return DeploymentState[state]
        except ValueError:
            return DeploymentState.UNKNOWN


@dataclass
class PendingModel:
    """A generic data model to be stored in the pending DB."""


@dataclass
class PendingUser(PendingModel):
    """Data model for storing information on a user pending approval."""

    unique_id: str
    sub: str
    iss: str
    email: Optional[str]
    full_name: Optional[str]
    username: str
    state: DeploymentState
    cmd: str
    infodict: dict


@dataclass
class PendingGroup(PendingModel):
    """Data model for storing groups of a user pending approval."""

    name: str
    state: DeploymentState
    cmd: str
    infodict: dict


@dataclass
class PendingMemberships(PendingModel):
    """Data model for storing group memberships of a user pending approval."""

    unique_id: str
    supplementary_groups: list
    removal_groups: list
    state: DeploymentState
    cmd: str
    infodict: dict



"""Generic backend User and Group, to be implemented by all backends."""

# vim: foldmethod=indent : tw=100
# pylint: disable=invalid-name, superfluous-parens
# pylint: disable=logging-fstring-interpolation, logging-not-lazy, logging-format-interpolation
# pylint: disable=raise-missing-from, missing-docstring, too-few-public-methods

from abc import ABC, abstractmethod
from ldf_adapter.backend.hooks import Hooks


class User(ABC, Hooks):
    """Manages the user object on the service."""

    def __init__(self, userinfo, **hooks):
        """
        Arguments:
        userinfo -- (type: UserInfo)
        """
        super().__init__(**hooks)

    @abstractmethod
    def exists(self):
        """Return whether the user exists on the service.

        If this returns True,  calling `create` should have no effect or raise an error.
        """
        return bool()

    @abstractmethod
    def name_taken(self, name):
        """Return whether a given username is already taken by another user on the service.

        Should return True if the name is not available for this user (even if it is available
        for other users for some reason)
        """
        return bool()

    @abstractmethod
    def create(self):
        """Create the user on the service.

        If the user already exists, do nothing or raise an error
        """
        pass

    @abstractmethod
    def create_tostring(self):
        """String representation of commands needed to deploy a new user.

        Returns:
            str: backend specific string containing user creation command(s)
        """
        pass

    # def create_fromstring(self, create_cmd: str):
    #     """Create a new user from string representation of commands needed to deploy a new user.
    #     Optional, only needed if the backend supports it.
    #
    #     Arguments:
    #         create_cmd str: backend specific string containing user creation command(s)
    #     """
    #     pass

    @abstractmethod
    def update(self):
        """Update all relevant information about the user on the service.

        If the user doesn't exists, behaviour is undefined.
        """
        pass

    @abstractmethod
    def delete(self):
        """Delete the user on the service.

        If the user doesn't exists, do nothing or raise an error.
        """
        pass

    @abstractmethod
    def mod(self, supplementary_groups=None):
        """Modify the user on the service.

        After this operation, the user will only be part of the provided groups.

        If the user doesn't exists, behaviour is undefined.

        Arguments:
        supplementary_groups -- A list of groups the user must be part of (type: list(Group))

        Returns:
        two lists of groups: the groups the user was added to and the groups the user was removed from
        """
        pass

    @abstractmethod
    def mod_tostring(self, supplementary_groups=None):
        """String representation of commands needed to modify a user to be added and removed from given groups.

        Arguments:
            supplementary_groups (list[Group], optional): the list of groups the user must be part of. Defaults to None.

        Returns:
            str: backend specific string containing all user modifications
        """
        pass

    # def mod_fromstring(self, mod_cmd: str):
    #     """Modify a user using string representation of commands needed to add & remove user to groups.
    #     Optional, only needed if the backend supports it.
    #
    #     Arguments:
    #         mod_cmd str: backend specific string containing all user modification command(s)
    #     """
    #     pass

    @abstractmethod
    def get_groups(self):
        """Get a list of names of all service groups that the user belongs to.

        If the user doesn't exist, behaviour is undefined.
        """
        pass

    @abstractmethod
    def install_ssh_keys(self):
        """Install users SSH keys on the service.

        No other SSH keys should be active after calling this function.

        If the user doesn't exists, behaviour is undefined.
        """
        pass

    @abstractmethod
    def uninstall_ssh_keys(self):
        """Uninstall the users SSH keys on the service.

        This must uninstall all SSH keys installed with `install_ssh_keys`. It may uninstall SSH
        keys installed by other means.

        If the user doesn't exists, behaviour is undefined.
        """
        pass

    @abstractmethod
    def get_username(self):
        """Return local username on the service.

        If the user doesn't exists, behaviour is undefined.
        """
        pass

    @abstractmethod
    def set_username(self, username):
        """Set local username on the service."""
        pass

    @abstractmethod
    def get_primary_group(self):
        """Check if a user exists based on unique_id and return the primary group name."""
        pass

    def is_suspended(self):
        """Optional, only if the backend supports it.
        Return whether the user was suspended (e.g. due to a security incident)"""
        return False

    def is_limited(self):
        """Optional, only if the backend supports it.
        Return whether the user has limited access"""
        return False

    def suspend(self):
        """Optional, only if the backend supports it.
        Suspends the user such that no access to the service is possible"""
        pass

    def resume(self):
        """Optional, only if the backend supports it.
        Restores the suspended user"""
        pass

    def limit(self):
        """Optional, only if the backend supports it.
        Limits the user's capabilities on the service (e.g. read-only access)"""
        pass

    def unlimit(self):
        """Optional, only if the backend supports it.
        Restores a user with limited access to full capabilities"""
        pass


class Group(ABC):
    """Manages the group object on the service."""

    def __init__(self, name):
        """
        Arguments:
        name -- The name of the group
        """
        pass

    @abstractmethod
    def exists(self):
        """Return whether the group already exists."""
        pass

    @abstractmethod
    def create(self):
        """Create the group on the service.

        If the group already exists, behaviour is undefined.
        """
        pass



import copy
from typing import Type
from os.path import basename, isfile, join
import glob
from dataclasses import dataclass

from ldf_adapter.results import FatalError


def dictdiff(old, new):
    """Return the difference between two dicts.

    Returns a dictionary mapping the keys from `old` and `new` to tuples `(old_value, new_value)`.
    A value of `None` in one of the elements of the tuple indicates that the key was not present in
    the respective dictionary.

    Does not distinguishing between missing keys and values of `None`.

    Arguments:
    old -- The first dictionary (type: dict)
    new -- The other dictionary (type: dict)
    """

    def _dictdiff(old, new):
        for k in new:
            if isinstance(new.get(k), dict) and isinstance(old.get(k), dict):
                subdiff = dict(_dictdiff(old[k], new[k]))
                if subdiff:
                    yield (k, subdiff)
            elif old.get(k) != new.get(k):
                yield (k, (old.get(k), new.get(k)))

    return dict(_dictdiff(old, new))


def log_dictdiff(diff, log_function=print, prefix=""):
    """Print the given dict-difference.

    Arguments:
    diff -- The dict diff as returned by `dictdiff` (type: dict)
    log_function -- The function to be used for printing (type: lambda str: None)
    """
    for k, v in diff.items():
        if isinstance(v, dict):
            log_dictdiff(v, log_function, "{}/".format(k))
        else:
            (old, new) = v
            if old:
                log_function(
                    "Updating {}{} from '{}' to '{}'".format(prefix, k, old, new)
                )
            else:
                log_function("Setting {}{} to '{}'".format(prefix, k, new))


def dictmerge(lhs, rhs):
    """Merge two dicts recusively."""
    res = copy.deepcopy(lhs)
    for k in rhs:
        if isinstance(rhs[k], dict) and k in res and isinstance(res[k], dict):
            res[k] = dictmerge(res[k], rhs[k])
        else:
            res[k] = rhs[k]

    return res


def to_bool(bool_str):
    """Convert a string to bool.
    Raise a FatalError if the string cannot be converted.
    """
    if bool_str.lower() in ["true", "yes", "yes, do as i say!"]:
        return True
    if bool_str.lower() in ["false", "no"]:
        return False
    raise FatalError(
        f"Error converting to bool: unrecognised boolean value {bool_str}."
    )


def to_int(int_str):
    """Convert a string to int.
    Raise a FatalError if the string cannot be converted.
    """
    try:
        return int(int_str)
    except ValueError:
        raise FatalError(
            f"Error converting to int: unrecognised integer value {int_str}."
        )


def to_list(list_str):
    """Convert a string containing comma-separated strings to list of strings.
    Raise a FatalError if the string cannot be converted.
    """
    try:
        return list(set(list_str.split()))
    except ValueError:
        raise FatalError(
            f"Error converting to list: unrecognised list value {list_str}."
        )


def dynamic_import(fq_module_name: str, class_name: str) -> Type:
    """Import a class dynamically.

    Args:
        fq_module_name: The fully qualified name of the module to import.
        class_name: The name of the class to import.
    Returns:
        The imported class.
    """
    module = __import__(fq_module_name, fromlist=[class_name])
    try:
        return getattr(module, class_name)
    except AttributeError:
        raise FatalError(
            message=f"Module {fq_module_name} does not contain class {class_name}"
        )


@dataclass
class BuilderConfig:
    fq_module_name: str
    class_name: str


class ObjectFactory:
    """Generic factory interface for creating all kinds of objects."""

    def __init__(self):
        self._builders = {}
        self._configs = {}

    def register_builder(self, key, fq_module_name, class_name):
        """Register a Builder based on a key value. This only saves the builder configuration."""
        self._configs[key] = BuilderConfig(fq_module_name, class_name)

    def _create_builder(self, key):
        """Create a Builder for a key value. The builder configuration needs to have been
        registered. Only at this point, the module containing the class is dynamically loaded.
        """
        config = self._configs.get(key)
        if config is None:
            raise FatalError(message=f"No builder registered for key {key}")
        object_type = dynamic_import(config.fq_module_name, config.class_name)
        self._builders[key] = ObjectBuilder(object_type)

    def _get_builder(self, key):
        """Return the Builder for a given key. Create it if it's registered but does not exist."""
        builder = self._builders.get(key)
        if not builder:
            self._create_builder(key)
            builder = self._builders.get(key)
            if not builder:
                raise FatalError(message=f"No builder created for key {key}")
        return builder

    def get(self, key, **kwargs):
        """Create and return the concrete object instance based on the key, with given arguments."""
        builder = self._get_builder(key)
        return builder(**kwargs)

    def get_type(self, key):
        """Return the type of the concrete object based on the key."""
        builder = self._get_builder(key)
        return builder._type


class ObjectBuilder:
    """Generic builder for creating all kinds of objects of given type."""

    def __init__(self, object_type: Type):
        self._instance = None
        self._type = object_type

    def __call__(self, **config):
        if not self._instance:
            self._instance = self._type(**config)
        return self._instance


def create_factory(dir_name: str, parent_module: str, class_name: str) -> ObjectFactory:
    """Create a factory for a given class name by registering object builders for
    all sub-modules in a given directory.

    E.g.
        create_factory(".../ldf_adapter/approval/db", "ldf_adapter.approval.db", "PendingDB")
    will register builders for all sub-modules in the directory ".../ldf_adapter/approval/db"
    and create a factory for the class "PendingDB" in the module "ldf_adapter.approval.db".
    The builders will be registered with the factory using the name of the sub-module as key.

    The sub-modules are not loaded at this time, rather on demand.

    Args:
        dir_name (str): name of directory containing modules that provide implementations of the class.
        parent_module (str): name of the parent module.
        class_name (str): name of class to provide objects for.

    Returns:
        ObjectFactory: factory for the given class.
    """
    factory = ObjectFactory()
    for file in glob.glob(join(dir_name, "*.py")):
        if (
            isfile(file)
            and not file.endswith("__init__.py")
            and not file.endswith("generic.py")
        ):
            module_name = basename(file)[:-3]
            fq_module_name = f"{parent_module}.{module_name}"
            factory.register_builder(module_name, fq_module_name, class_name)
    return factory



from abc import ABC, abstractmethod

from string import Template
from enum import Enum
from pathlib import Path


class NotificationType(Enum):
    NONE = 0
    ADMIN_DEPLOY = 1
    ADMIN_DEPLOY_UPDATE = 2
    ADMIN_UPDATE = 3
    ADMIN_TEST = 4
    USER_DEPLOY = 5
    USER_DEPLOY_UPDATE = 6
    USER_UPDATE = 7
    UNKNOWN = -1


class NotificationTemplate:
    def __init__(self, template_body: str, template_id: str = ""):
        """Initialises notification template.

        Args:
            template_body (str): body of the template
            template_id (str, optional): id of the template. Defaults to "".
        """
        self.template_body = template_body
        self.template_id = template_id

    def get_variables(self) -> list:
        """Returns list of variables in the template."""
        return [
            var for var in Template(self.template_body).pattern.split("$") if var != ""
        ]

    def fill(self, **kwargs):
        """Fills template with given variables.

        Args:
            **kwargs: variables to fill template with.
        """
        # return Template(self.template_body).substitute(**kwargs)
        return Template(self.template_body).safe_substitute(**kwargs)

    @classmethod
    def load(cls, template_file: Path) -> "NotificationTemplate":
        """Loads notification template from file.

        Args:
            template_file (Path): path to the template file.
        """
        with open(template_file, "r") as f:
            return cls(
                template_body=f.read(),
                template_id=Path(template_file).name,
            )


class Notifier(ABC):
    """Generic class for sending notifications."""

    @abstractmethod
    def notify(self, notification_type: NotificationType, **kwargs: dict) -> bool:
        """Sends out a notification. Arguments are not specified here, but are passed to the
        concrete implementation.
        Returns True if notification was sent, False otherwise.

        To be implemented for all notification providers.
        """
        pass

    @abstractmethod
    def test(self):
        """Send test notification to admin to make sure the configured set-up works.

        To be implemented for all notification providers.
        """
        pass



import os
import subprocess

from ldf_adapter.logsetup import logger


class Hooks:
    def __init__(self, **hooks):
        for k, v in hooks.items():
            setattr(self, k, v)
            logger.debug(f"Hook {k} set to {v}")

    def execute(self, hook, *hook_args):
        if not hasattr(self, hook):
            logger.error(f"Hook '{hook}' not implemented")
            return

        hook_script = getattr(self, hook)
        logger.debug(f"Executing hook for {hook}: {hook_script}")

        if not os.path.isfile(hook_script):
            logger.error(f"Hook for {hook}: {hook_script} does not exist, skipping.")
            return
        command = [hook_script, *hook_args]
        if hook_script.endswith(".sh"):
            command.insert(0, "bash")
        elif hook_script.endswith(".py"):
            command.insert(0, "python3")
        else:
            logger.error(
                f"Hook for {hook}: {hook_script} is not a bash or python script, skipping."
            )
            return
        try:
            logger.info(f"Running hook for {hook}: {hook_script}")
            subprocess.run(
                command,
                check=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
        except subprocess.CalledProcessError as e:
            logger.error(
                f"Error running hook for {hook}: {hook_script}: {e}. Skipping."
            )
            return



name = "ldf_adapter"
# vim: tw=100 foldmethod=indent
# pylint: disable=invalid-name, superfluous-parens
# pylint: disable=redefined-outer-name, logging-not-lazy, logging-format-interpolation, logging-fstring-interpolation
# pylint: disable=missing-docstring, trailing-whitespace, trailing-newlines, too-few-public-methods

import logging
from feudal_globalconfig import globalconfig

import regex

from ldf_adapter import backend
from ldf_adapter.config import CONFIG
from ldf_adapter.results import (
    Deployed,
    Failure,
    NotDeployed,
    Rejection,
    Question,
    Status,
    FatalError,
)
from ldf_adapter.name_generators import NameGenerator
from ldf_adapter.userinfo import UserInfo
from ldf_adapter.approval import PendingDeployment

logger = logging.getLogger(__name__)


class User:
    """Represents a user, abstracting from the concrete service.

    An abstract User is backed by a service_user and is associated with a set of groups (backed by
    service_groups).

    A user is usually identified on the service not by a username but by `self.data.unique_id` (see
    __init__ for details).
    """

    def __init__(self, data):
        """
        Arguments:
        data -- Information about the user (type: UserInfo or dict)

        Relevant config:
        ldf_adapter.backend -- The name of the backend. See function the `backend` for possible values
        ldf_adapter.primary_group -- The primary group of the user. If empty, one from the
          supplementary groups will be used. If there are multiple, a question will be raised.

        Words of warning: Since the service_user and service_groups are
        backend specific, their structures differ from backend to backend.

        Direct access to them from this __init__ is highly illegal (unless specified)
        Instead: Use self.data
        """
        # Info Display Hack
        if CONFIG.verbose_info_plugin.active:
            try:
                if data["state_target"] == "deployed":
                    import json
                    import os
                    import stat

                    filename = CONFIG.verbose_info_plugin.filename
                    dirname = CONFIG.verbose_info_plugin.dirname
                    try:
                        os.mkdir(dirname)
                        os.chmod(dirname, 0o0777)
                    except:
                        pass
                    with open(filename, "w", encoding="utf-8") as f:
                        json.dump(
                            data,
                            f,
                            ensure_ascii=False,
                            sort_keys=True,
                            indent=4,
                            separators=(",", ": "),
                        )
                    os.chmod(filename, 0o0666)
                    # os.chmod(filename, stat.S_IREAD || stat.S_IWRITE ||stat.S_IRGRP || stat.S_IWGRP
                    #         || stat.S_IROTH || stat.S_IWOTH)
            except Exception as e:
                logger.error(f"Got an exception in uncritical code: {e}")

        hooks = {}
        backend_config = CONFIG.backend.__getattribute__(CONFIG.ldf_adapter.backend)
        if (
            hasattr(backend_config, "post_create_script")
            and backend_config.post_create_script
        ):
            hooks["post_create"] = backend_config.post_create_script
        else:
            logger.debug(
                f"post_create_script not supported for backend {CONFIG.ldf_adapter.backend}"
            )

        # Proceed as normal
        self.data = data if isinstance(data, UserInfo) else UserInfo(data)
        self.service_user = backend.User(self.data, **hooks)
        self.service_groups = [backend.Group(grp) for grp in self.data.groups]

        # add additional groups from config
        self.additional_groups = [
            backend.Group(grp)
            for grp in CONFIG.ldf_adapter.additional_groups
            if grp not in self.data.groups
        ]

        if CONFIG.ldf_adapter.backend_supports_preferring_existing_user:
            logger.debug("trying to update user from existing")
            if self.service_user.exists():
                self.update_username_from_existing()

        if CONFIG.approval.enabled:
            # initialise pending deployment and notification system
            self.pending_deployment = PendingDeployment(userinfo=self.data)

        # apply fixes to group names for unix backend
        if CONFIG.ldf_adapter.backend == "local_unix" and hasattr(
            backend.User, "fix_group_names"
        ):
            for grp in self.service_groups + self.additional_groups:
                grp.fix_group_names()

    def assurance_verifier(self):
        """Produce a suitable function to check if a user is allowed.

        Relevant config:
        assurance.prefix -- The common prefix of all relative assurance claims
        assurance.require -- The boolean expression to be parsed, according to the following
          grammer: `E -> E "&" E | E "|" E | "(" E ")" | string`, where `&` binds stronger than `|`.
          Strings are assurance claims interpreted absolute (if they start with `"http[s]://"`) or
          relative to `assurance.prefix`. The strings "+" and "*" are interpreted specially: "+"
          means "any assurance claim", while "*" means "any claim, or no claim at all". They thus
          differ in their treatment of users without any claims.

        Returns:
        A function taking a set of assurance claims, interpreted absolute. The function returns
        `True`, if the claims satisfy the configured expression (`"assurance.require"`), `False`
        otherwise.
        """
        prefix = CONFIG.assurance.prefix
        prefix = prefix.rstrip("/") + "/"

        tokens = regex.findall("&|\||\(|\)|[^\s()&|]+", CONFIG.assurance.require)
        #  tokens = regex.findall(r"&|\||\(|\)|[^\s()&|]+", CONFIG.assurance.require)
        #  logger.info(F"assurance tokens: {tokens}")

        # We use a simple recursive descent parser to parse parenthesied expressions of strings,
        # composed with '&' (konjunction) and '|' (disjunction). The usual precedence rules apply.
        #
        # Instead of building an AST, we build a tree of nested lambdas, which takes a collection of
        # assurance claims and checks if they satisfy the configured expression

        #  EXPR -> DISJ 'EOF'
        def parse_expr(seq):
            expr = parse_disjunction(seq)
            if len(seq) > 0:
                raise ValueError("Reached end of input while parsing")
            return expr

        #  DISJ -> KONJ DISJ2
        def parse_disjunction(seq):
            lhs = parse_konjunction(seq)
            return parse_disjunction2(seq, lhs)

        #  DISJ2 -> ""
        #        -> "|" KONJ DISJ2
        def parse_disjunction2(seq, lhs):
            if len(seq) > 0 and seq[0] == "|":
                seq.pop(0)
                rhs = parse_konjunction(seq)
                expr = lambda values: lhs(values) or rhs(values)
                return parse_disjunction2(seq, expr)
            else:
                return lhs

        #  KONJ -> PRIMARY KONJ2
        def parse_konjunction(seq):
            lhs = parse_primary(seq)
            return parse_konjunction2(seq, lhs)

        #  KONJ2 -> ""
        #        -> "&" PRIMARY
        def parse_konjunction2(seq, lhs):
            if len(seq) > 0 and seq[0] == "&":
                seq.pop(0)
                rhs = parse_primary(seq)
                expr = lambda values: lhs(values) and rhs(values)
                return parse_konjunction2(seq, expr)
            else:
                return lhs

        #  PRIMARY -> "(" DISJ ")"
        #          -> ASSURANCE
        def parse_primary(seq):
            if len(seq) > 0 and seq[0] == "(":
                seq.pop(0)
                subexpr = parse_disjunction(seq)
                if len(seq) > 0 and seq.pop(0) != ")":
                    raise ValueError("Missing ')' while parsing")
                return subexpr
            else:
                return parse_assurance(seq)

        #  ASSURANCE -> string
        #            -> "*"
        #            -> "+"
        def parse_assurance(seq):
            value = seq.pop(0)
            if value == "+":
                return lambda values: len(values) > 0
            elif value == "*":
                return lambda values: True
            else:
                value = value if regex.match("https?://", value) else prefix + value
                return lambda values: value in values

        return parse_expr(tokens)

    def reach_state(self, target):
        """Attempt to put the user into the desired state on the configured service.

        Arguments:
        target -- The desired state. One of 'deployed' and 'not_deployed'.
        user -- The user to be deployed/undeployed (type: User)
        """

        username = "not yet assigned"
        if self.service_user.exists():
            username = self.service_user.get_username()
        try:
            logger.info(
                f"Incoming request to reach '{target}' for user with email: '{self.data.email}' ({self.data.unique_id}) username: {username}"
            )
        except AttributeError:
            logger.info(
                f"Incoming request to reach '{target}' for user with name: '{self.data.full_name}' ({self.data.unique_id}) username: {username}"
            )

        if target == "deployed":
            if CONFIG.assurance.skip:
                logger.warning(
                    "Assurance checking is disabled: Users with ANY assurance will be authorised"
                )
            elif not self.assurance_verifier()(self.data.assurance):
                raise Rejection(
                    message="Your assurance level is insufficient to access this resource"
                )

            logger.debug(f"User comes with these groups")
            for g in self.service_groups:
                logger.debug(f"    {g.name}")

            return self.deploy()
        elif target == "not_deployed":
            if not CONFIG.assurance.skip and not self.assurance_verifier()(
                self.data.assurance
            ):
                if not CONFIG.assurance.verified_undeploy:
                    logger.warning(
                        "Assurance level is insufficient. Undeploying anyway."
                    )
                else:
                    raise Rejection(
                        message="Your assurance level is insufficient to access this resource"
                    )
            return self.undeploy()
        elif target == "get_status":
            return self.get_status()
        elif target == "resumed":
            return self.resume()
        elif target == "suspended":
            return self.suspend()
        elif target == "limited":
            return self.limit()
        elif target == "accepted":
            return self.accept()
        elif target == "rejected":
            return self.reject()
        elif target == "test":
            return self.test()
        else:
            raise ValueError(f"Invalid target state: {target}")

    def deploy(self):
        """Deploy the user.

        Ensure that the user exists, is a member in the right groups (and only in those groups)
        and has the correct credentials installed.

        Return a Deployed result, with a message describing what was done.
        """
        if CONFIG.approval.enabled and self.pending_deployment.is_rejected():
            return Status(
                state="rejected",
                message="User deployment was rejected. No new deployment request will be sent.",
            )

        self.ensure_groups_exist()
        was_created = self.ensure_exists()
        new_memberships, removed_memberships = self.ensure_group_memberships()
        new_credentials = self.ensure_credentials_active()
        if was_created:
            self.service_user.execute("post_create", self.service_user.get_username())

        if CONFIG.approval.enabled:
            self.pending_deployment.notify()
            # if user existed
            if self.service_user.exists():
                what_changed = "User already existed"
                if new_memberships != [] or removed_memberships != []:
                    what_changed += ", but groups changed. A request for updating user groups was submitted for approval"
                what_changed += "."
                return Deployed(message=what_changed, credentials=self.credentials)
            # user is pending
            if was_created:
                what_changed = "Request for deployment was submitted for approval."
            elif new_memberships != [] or removed_memberships != []:
                what_changed = (
                    "Updated request for deployment was submitted for approval."
                )
            else:
                what_changed = "Request for deployment was already submitted for approval. No new request was sent."
            return Status(state="pending", message=what_changed)
        else:
            what_changed = "User was created" if was_created else "User already existed"
            if new_memberships != []:
                what_changed += " and was added to groups {}".format(
                    ",".join(new_memberships)
                )
            if removed_memberships != []:
                what_changed += " and was removed from groups {}".format(
                    ",".join(removed_memberships)
                )
            what_changed += "."
            if new_credentials:
                what_changed += " Credentials {} were activated.".format(
                    ",".join(new_credentials)
                )

            return Deployed(credentials=self.credentials, message=what_changed)

    def undeploy(self):
        """Ensure that the user dosen't exist.

        Return a NotDeployed result with a message saying if the user previously existed.
        """
        username = self.service_user.get_username()
        was_removed = self.ensure_doesnt_exist()

        what_changed = ""
        if was_removed:
            what_changed += f"User '{username} ({self.data.unique_id})' was removed."
        else:
            what_changed += (
                f"No user for '{self.data.unique_id}' existed. "
                + f"User '{username}' was not changed"
            )

        return NotDeployed(message=what_changed)

    def suspend(self):
        """Ensure that the user is suspended.

        Return a Status result with a message describing what was done.
        """
        was_suspended = self.ensure_suspended()
        what_changed = ""
        if was_suspended:
            # FIXME: I'm not sure if this ought to be self.data.username
            what_changed += f"User '{self.data.unique_id}' was suspended."
            state = "suspended"
        else:
            state = self.get_status().state
            what_changed += (
                f"Suspending user '{self.data.unique_id}' was not possible from the '{state}' state. "
                + f"User was not changed."
            )
        return Status(state, message=what_changed)

    def resume(self):
        """Ensure that a suspended user is active again in state 'deployed'.

        Return a Status result with a message describing what was done.
        """
        was_resumed = self.ensure_resumed()
        what_changed = ""
        state = self.get_status().state
        if was_resumed:
            # FIXME: I'm not sure if this ought to be self.data.username
            what_changed += f"User '{self.data.unique_id}' was resumed."
        else:
            what_changed += (
                f"Resuming user '{self.data.unique_id}' was not possible from the '{state}' state. "
                + f"User was not changed."
            )
        return Status(state, message=what_changed)

    def limit(self):
        """Ensure that a user has limited access.

        Return a Status result with a message describing what was done.
        """
        was_limited = self.ensure_limited()
        what_changed = ""
        if was_limited:
            # FIXME: I'm not sure if this ought to be self.data.username
            what_changed += f"User '{self.data.unique_id}' was limited."
            state = "limited"
        else:
            state = self.get_status().state
            what_changed += (
                f"Limiting user '{self.data.unique_id}' was not possible from the '{state}' state. "
                + f"User was not changed."
            )
        return Status(state, message=what_changed)

    def unlimit(self):
        """Ensure that a limited user is active again in state 'deployed'.

        Return a Status result with a message describing what was done.
        """
        was_unlimited = self.ensure_unlimited()
        what_changed = ""
        if was_unlimited:
            # FIXME: I'm not sure if this ought to be self.data.username
            what_changed += f"User '{self.data.unique_id}' was unlimited."
            state = "deployed"
        else:
            state = self.get_status().state
            what_changed += (
                f"Resuming user '{self.data.unique_id}' was not possible from the '{state}' state. "
                + f"User was not changed."
            )
        return Status(state, message=what_changed)

    def accept(self):
        """Ensure that a pending request is accepted and the user is in state 'deployed'.

        Return a Status result with a message describing what was done.
        """
        if (
            self.pending_deployment.is_pending()
            or self.pending_deployment.mod_pending()
        ):
            if not hasattr(backend.User, "create_fromstring"):
                raise Failure(
                    message=(
                        "Backend does not support automatic user creation from string."
                        "Follow the instructions in the notification you received to manually create the user."
                    )
                )
            if not hasattr(backend.Group, "create_fromstring"):
                raise Failure(
                    message=(
                        "Backend does not support automatic group creation from string."
                        "Follow the instructions in the notification you received to manually create the group(s)."
                    )
                )
            if not hasattr(backend.User, "mod_fromstring"):
                raise Failure(
                    message=(
                        "Backend does not support automatic user modification from string."
                        "Follow the instructions in the notification you received to manually add the user to groups."
                    )
                )
            self.pending_deployment.accept()
            return Deployed(
                credentials=self.credentials, message="User request was accepted."
            )
        elif self.pending_deployment.is_rejected():
            what_changed = "User request was already rejected, cannot accept it."
            logger.debug(what_changed)
            return Status(state="rejected", message=what_changed)
        else:
            what_changed = f"No pending request for user {self.data.unique_id} exists."
            logger.debug(what_changed)
            return Status(state=self.get_status().state, message=what_changed)

    def reject(self):
        """Ensure that a pending request is rejected and the user is in state 'not_deployed'.

        Return a Status result with a message describing what was done.
        """
        if (
            self.pending_deployment.is_pending()
            or self.pending_deployment.mod_pending()
        ):
            self.pending_deployment.reject()
            what_changed = "User deployment request was rejected."
            logger.debug(what_changed)
            return Status(state="rejected", message=what_changed)
        else:
            what_changed = f"No pending request for user {self.data.unique_id} exists."
            logger.debug(what_changed)
            return NotDeployed(message=what_changed)

    def get_status(self):
        """
        Return the current status (that he has in the underlying local user management system)
        User can have these status:
        +--------------+-----------------------------------------------------------------+-----------------+
        | Status       | Comment                                                         | Backend support |
        +--------------+-----------------------------------------------------------------+-----------------+
        +--------------+-----------------------------------------------------------------+-----------------+
        | deployed     | There is an account for the user identified by unique_id        | Mandatory       |
        +--------------+-----------------------------------------------------------------+-----------------+
        | not deployed | There is no account for the user identified by unique_id        | Mandatory       |
        |              | We have no information if there has ever been an account        |                 |
        +--------------+-----------------------------------------------------------------+-----------------+
        | rejected     | This might not be supportable; Depends on the backend           | Optional        |
        +--------------+-----------------------------------------------------------------+-----------------+
        | suspended    | The user with unique_id has been suspended                      | Optional        |
        +--------------+-----------------------------------------------------------------+-----------------+
        | pending      | The creation of the user is pending                             | Optional        |
        +--------------+-----------------------------------------------------------------+-----------------+
        | limited      | The user was limited, typically after being idle for some time  | Optional        |
        +--------------+-----------------------------------------------------------------+-----------------+
        | unknown      | We don't know the status, but at least the user is not deployed | Mandatory       |
        +--------------+-----------------------------------------------------------------+-----------------+
        """

        msg = "No message"
        try:
            if not self.service_user.exists():
                if CONFIG.approval.enabled:
                    if self.pending_deployment.is_pending():
                        return Status(
                            "pending", message="User deployment is pending approval."
                        )
                    if self.pending_deployment.is_rejected():
                        return Status(
                            "rejected", message="User deployment was rejected."
                        )
                return Status("not_deployed", message=msg)
            msg = f"username {self.service_user.get_username()}"
            if hasattr(self.service_user, "is_suspended"):
                if self.service_user.is_suspended():
                    return Status("suspended", message=msg)
            if hasattr(self.service_user, "is_limited"):
                if self.service_user.is_limited():
                    return Status("limited", message=msg)
            return Status("deployed", message=msg)
        except Exception as e:
            logger.error(f"User {self.data.unique_id} is in an undefined state.: {e}")
            return Status("unknown", message=msg)

    def ensure_exists(self):
        """Ensure that the user exists on the service.

        If the username is already taken on the service, raise a questionaire for a new one. See
        UserInfo.username for details.

        Also ensure that all info about the user is up to date on the service. This is done
        independently of creating the user, so that the user is updated even if they already existed.

        Return True, if the user didn't exist before and there was no pending request for its creation.
        """
        logger.debug(f"Ensuring a local user mapping for {self.data.unique_id} exits")

        is_new_user = (
            not self.service_user.exists()
            if not CONFIG.approval.enabled
            else not self.service_user.exists() and not self.pending_deployment.exists()
        )

        unique_id = self.data.unique_id
        if is_new_user:
            username = self.data.username
            primary_group_name = self.service_user.primary_group.name

            # Raise question in case of existing username in case we're interactive
            if CONFIG.ldf_adapter.interactive:  # interactive
                logger.debug("interactive mode")
                if self.service_user.name_taken(username):
                    logger.info(
                        f'Username "{username}" is already taken, asking user to pick a new one'
                    )
                    raise Question(
                        name="username",
                        text=f'Username "{username}" already taken on this service. Please enter another one.',
                    )

            else:  # non-interactive
                logger.debug("noninteractive mode")
                username_mode = CONFIG.username_generator.mode
                logger.debug(f"username_mode: {username_mode}")
                pool_prefix = (
                    CONFIG.username_generator.pool_prefix or primary_group_name
                )

                name_generator = NameGenerator(
                    username_mode, userinfo=self.data, pool_prefix=pool_prefix
                )
                proposed_name = name_generator.suggest_name()
                logger.debug(f"initially proposed_name: {proposed_name}")

                while proposed_name is not None and (
                    self.service_user.name_taken(proposed_name)
                    or (
                        CONFIG.approval.enabled
                        and self.pending_deployment.name_taken(proposed_name)
                    )
                ):
                    proposed_name = name_generator.suggest_name()
                if proposed_name is None:
                    raise Rejection(
                        message=f"I cannot create usernames. "
                        f"The list of tried ones is: {', '.join(name_generator.tried_names())}."
                    )
                self.service_user.set_username(proposed_name)

                logger.info(f"Chose username '{proposed_name}' for {unique_id}")

            logger.debug(f"Primary Group Name: {primary_group_name}")
            logger.debug(f"Primary Group from userinfo: {self.data.primary_group}")

            # Sanity check to ensure user has a primary group:
            if primary_group_name is None:
                config_file_name = globalconfig.info["config_files_read"]
                message = (
                    'User is not member of any group, and neither a "primary_group", nor '
                    f'a "fallback_group" have been defined in the config file {config_file_name}'
                )
                logger.error(message)
                raise FatalError(message=message)

            if CONFIG.approval.enabled:
                self.pending_deployment.create_user(self.service_user)
            else:
                self.service_user.create()
        elif (
            CONFIG.approval.enabled and self.pending_deployment.exists()
        ):  # the user is pending
            username = self.pending_deployment.username
            self.service_user.set_username(username)
            logger.info(
                f"Username {username} already assigned to '{self.data.unique_id}' in pending request."
            )
        else:  # The user exists
            # Update service_user.name if unique_id already points to a username:
            username = self.service_user.get_username()
            logger.info(f"User {username} for '{self.data.unique_id}' already exists.")

        logger.debug(f"This is a new user: {is_new_user}")

        if not CONFIG.approval.enabled:
            self.service_user.update()
        return is_new_user

    def update_username_from_existing(self):
        """Update self.service_user.name, if a user with matching
        unique_id can be found, and if the backend implements 'set_username'
        """
        try:
            existing_username = self.service_user.get_username()
            if existing_username is not None:
                if hasattr(self.service_user, "set_username"):
                    logger.debug(
                        f"Setting username to {existing_username} ({self.data.unique_id})"
                    )
                    if hasattr(self.service_user, "set_prefixed_username"):
                        logger.debug("calling set_prefixed_username")
                        self.service_user.set_prefixed_username(existing_username)
                    else:
                        logger.debug("calling set_username")
                        self.service_user.set_username(existing_username)
                logger.debug(f"Found an existing username: {existing_username}")
        except AttributeError:
            # the currently used service_user class has to method get_username
            existing_username = None

    def ensure_doesnt_exist(self):
        """Ensure that the user doesn't exist.

        Before deleting them, uninstall all SSH keys, to be sure that they are really gone.

        Return True, if the user existed before.
        """
        if self.service_user.exists():
            self.service_user.username = self.service_user.get_username()
            logger.info(
                f"Deleting user '{self.service_user.username}' ({self.data.unique_id})"
            )
            # bwIDM requires prior removal of the user, because ssh-key removal triggers an
            # asyncronous process. If user is removed during that, the user might be only partially
            # removed...
            if CONFIG.ldf_adapter.backend == "bwidm":
                self.service_user.delete()
                self.service_user.uninstall_ssh_keys()
            else:
                self.service_user.uninstall_ssh_keys()
                self.service_user.delete()
            return True
        elif CONFIG.approval.enabled and self.pending_deployment.exists():
            logger.info(
                f"No user existed for {self.data.unique_id}, but cleaned up lingering pending/rejected entry for this user."
            )
            self.pending_deployment.remove_data()
            return False
        else:
            logger.info(f"No user existed for {self.data.unique_id}. Nothing to do.")
            return False

    def ensure_suspended(self):
        """Ensure that a user is suspended.
        Return True if the user has been suspended.
        """
        status = self.get_status()
        if status.state in ["deployed", "limited"]:
            if hasattr(self.service_user, "suspend"):
                self.service_user.suspend()
                return True
        logger.debug(
            f"User {self.data.unique_id} in state {status.state}. Suspending not allowed."
        )
        return False

    def ensure_limited(self):
        """Ensure that a user has limited access.
        Return True if setting the user is limited.
        """
        status = self.get_status()
        if status.state == "deployed":
            if hasattr(self.service_user, "limit"):
                self.service_user.limit()
                return True
        logger.debug(
            f"User {self.data.unique_id} in state {status.state}. Limiting not allowed."
        )
        return False

    def ensure_resumed(self):
        """Ensure that a user is resumed.
        Return True is the user
        """
        status = self.get_status()
        if status.state == "suspended":
            if hasattr(self.service_user, "resume"):
                self.service_user.resume()
                return True
        logger.debug(
            f"User {self.data.unique_id} in state {status.state}. Resuming not allowed."
        )
        return False

    def ensure_unlimited(self):
        """Ensure that a user is not limited anymore.
        Return True is the user
        """
        status = self.get_status()
        if status.state == "limited":
            if hasattr(self.service_user, "unlimit"):
                self.service_user.unlimit()
                return True
        logger.debug(
            f"User {self.data.unique_id} in state {status.state}. Unlimit not allowed."
        )
        return False

    def ensure_groups_exist(self):
        """Ensure that all the necessary groups exist.

        Create the groups on the service, if necessary.
        Return the names of the groups that were created (or requested to be created).
        """
        group_list = self.service_groups + self.additional_groups
        if (
            self.data.primary_group
            not in self.data.groups + CONFIG.ldf_adapter.additional_groups
        ):
            group_list.append(self.service_user.primary_group)

        new_groups = []
        for group in filter(
            lambda grp: not grp.exists() and grp.name is not None, group_list
        ):
            logger.info("Creating group '{}'".format(group.name))
            if CONFIG.approval.enabled:
                if self.pending_deployment.create_group(group):
                    new_groups.append(group.name)
            else:
                group.create()
                new_groups.append(group.name)
        return new_groups

    def ensure_group_memberships(self):
        """Ensure that the user is a member of all the groups in self.service_groups and self.additional_groups.

        Return two lists:
        - the names of all groups the user was added to
        - the names of all groups the user was removed from
        """
        username = self.service_user.get_username()

        group_list = self.service_groups + self.additional_groups
        if (
            self.data.primary_group
            not in self.data.groups + CONFIG.ldf_adapter.additional_groups
        ):
            group_list.append(self.service_user.primary_group)

        logger.info(
            f"Ensuring user '{username}' ({self.data.unique_id}) is member of these groups: "
            f"{[grp.name for grp in group_list]}"
        )

        if CONFIG.approval.enabled:
            groups_added, groups_removed = self.pending_deployment.mod(
                service_user=self.service_user, supplementary_groups=group_list
            )
            logger.info(
                f"User '{username}' has to be added to the following groups: {groups_added}"
            )
            logger.info(
                f"User '{username}' has to be removed from the following groups: {groups_removed}"
            )
            return groups_added, groups_removed
        else:
            groups_added, groups_removed = self.service_user.mod(
                supplementary_groups=group_list
            )
            logger.info(
                f"User '{username}' was added to the following groups: {groups_added}"
            )
            logger.info(
                f"User '{username}' was removed from the following groups: {groups_removed}"
            )
            return groups_added, groups_removed

    def ensure_credentials_active(self):
        """Install all SSH Keys on the service.

        Return a list of the names/ids of all the keys now active.
        """
        if not CONFIG.approval.enabled:
            self.service_user.install_ssh_keys()
            return ["ssh:{name}/{id}".format(**key) for key in self.data.ssh_keys]
        return []

    @property
    def credentials(self):
        """The Credentials displayed to the user.

        Simply merges all the credentials provided by the service_user with those
        configured in the config file.

        See Deployed.__init__ for details on how this value is used.

        Relevant config:
        login_info -- Everything in this section is merged into the credentials dictionary.
        """
        ssh_user = self.service_user.get_username()
        commandline = "ssh {}@{}".format(ssh_user, CONFIG.login_info.ssh_host)
        return {
            **CONFIG.login_info.to_dict(),
            "ssh_user": ssh_user,
            "commandline": commandline,
        }

    def test(self):
        """Various tests for the admin to check the feudal configuration.

        Test that the notification system is configured correctly if the approval workflow is
        enabled by sending a test notification to the admin.
        """
        if CONFIG.approval.enabled:
            self.pending_deployment.test_notifier()
            msg = "Notification sent successfully."
        else:
            msg = "Approval workflow is not enabled. No notification message was sent."
        logger.info(msg)
        return Status(state="test", message=msg)



from trycourier import Courier

from ldf_adapter.results import FatalError
from ldf_adapter.config import CONFIG
from ldf_adapter.notifier import generic


class Notifier(generic.Notifier):
    """(WIP) Implementation of a Courier notifier for deployment requests."""

    def __init__(self) -> None:
        """Initialises Courier notifier."""
        if CONFIG.notifier.courier is None:
            raise FatalError("Courier notifier is not configured")
        self.client = Courier(auth_token=CONFIG.notifier.courier.api_key)
        self.ssh_host = CONFIG.login_info.ssh_host

    def notify(
        self,
        notification_type: generic.NotificationType,
        send_to: str,
        template_id: str,
        data: dict,
        **_ignored: dict
    ):
        """Notifies admin of request of given type.

        Args:
            notification_type (NotificationType): type of notification
            send_to (str): email address to send notification to
            template_id (str): template ID of configured email template in courier
            data (dict): data to use in email template
        Returns:
            bool: True if notification was sent, False otherwise
        """
        self.client.send_message(
            message={
                "to": {
                    "email": send_to,
                },
                "template_id": template_id,
                "data": data,
            }
        )

    def test(self):
        return super().test()



import pytest

from ldf_adapter.eduperson import Entitlement


@pytest.mark.parametrize(
    "raw,group",
    [
        (
            "urn:nid:namespace:group:group:subgroup1:subgroup2:role=admin#authority",
            "group",
        ),
        (
            "urn:geant:h-df.de:group:IMK-TRO-EWCC#login.helmholtz-data-federation.de",
            "IMK-TRO-EWCC",
        ),
        (
            "urn:geant:h-df.de:group:MyExampleColab#login.helmholtz-data-federation.de",
            "MyExampleColab",
        ),
        (
            "urn:geant:h-df.de:group:wlcg-test#login.helmholtz-data-federation.de",
            "wlcg-test",
        ),
        ("urn:geant:h-df.de:group:HDF#login.helmholtz-data-federation.de", "HDF"),
    ],
)
def test_entitlement_group(raw, group):
    ent = Entitlement(raw)
    assert ent.group == group


@pytest.mark.parametrize(
    "raw,subgroup",
    [
        (
            "urn:nid:namespace:group:group:subgroup1:subgroup2:role=admin#authority",
            ["subgroup1", "subgroup2"],
        ),
        ("urn:nid:namespace:group:group:subgroup:role=admin#authority", ["subgroup"]),
        ("urn:nid:namespace:group:group:role=admin#authority", []),
    ],
)
def test_entitlement_subgroup(raw, subgroup):
    ent = Entitlement(raw)
    assert ent.subgroups == subgroup


@pytest.mark.parametrize(
    "raw,full_namespace",
    [
        (
            "urn:nid:namespace:group:group:subgroup1:subgroup2:role=admin#authority",
            ["namespace"],
        ),
        (
            "urn:nid:ns1:ns2:ns3:group:group:subgroup1:subgroup2:role=admin#authority",
            ["ns1", "ns2", "ns3"],
        ),
        (
            "urn:geant:h-df.de:group:IMK-TRO-EWCC#login.helmholtz-data-federation.de",
            ["h-df.de"],
        ),
    ],
)
def test_entitlement_full_namespace(raw, full_namespace):
    ent = Entitlement(raw)
    assert ent.full_namespace == full_namespace


@pytest.mark.parametrize(
    "raw,namespace_id",
    [
        (
            "urn:nid:namespace:group:group:subgroup1:subgroup2:role=admin#authority",
            "nid",
        ),
        (
            "urn:nid:ns1:ns2:ns3:group:group:subgroup1:subgroup2:role=admin#authority",
            "nid",
        ),
        (
            "urn:geant:h-df.de:group:IMK-TRO-EWCC#login.helmholtz-data-federation.de",
            "geant",
        ),
    ],
)
def test_entitlement_namespace_id(raw, namespace_id):
    ent = Entitlement(raw)
    assert ent.namespace_id == namespace_id


@pytest.mark.parametrize(
    "raw,delegated_namespace",
    [
        (
            "urn:nid:namespace:group:group:subgroup1:subgroup2:role=admin#authority",
            "namespace",
        ),
        (
            "urn:nid:ns1:ns2:ns3:group:group:subgroup1:subgroup2:role=admin#authority",
            "ns1",
        ),
        (
            "urn:geant:h-df.de:group:IMK-TRO-EWCC#login.helmholtz-data-federation.de",
            "h-df.de",
        ),
    ],
)
def test_entitlement_delegated_namespace(raw, delegated_namespace):
    ent = Entitlement(raw)
    assert ent.delegated_namespace == delegated_namespace


@pytest.mark.parametrize(
    "raw,subnamespaces",
    [
        ("urn:nid:namespace:group:group:subgroup1:subgroup2:role=admin#authority", []),
        (
            "urn:nid:ns1:ns2:ns3:group:group:subgroup1:subgroup2:role=admin#authority",
            ["ns2", "ns3"],
        ),
        ("urn:geant:h-df.de:group:IMK-TRO-EWCC#login.helmholtz-data-federation.de", []),
    ],
)
def test_entitlement_subnamespaces(raw, subnamespaces):
    ent = Entitlement(raw)
    assert ent.subnamespaces == subnamespaces


@pytest.mark.parametrize(
    "raw,role",
    [
        (
            "urn:nid:namespace:group:group:subgroup1:subgroup2:role=admin#authority",
            "admin",
        ),
        (
            "urn:nid:namespace:group:group:subgroup1:subgroup2:role=user#authority",
            "user",
        ),
        (
            "urn:geant:h-df.de:group:IMK-TRO-EWCC#login.helmholtz-data-federation.de",
            None,
        ),
    ],
)
def test_entitlement_role(raw, role):
    ent = Entitlement(raw)
    assert ent.role == role


@pytest.mark.parametrize(
    "raw,group_authority",
    [
        (
            "urn:nid:namespace:group:group:subgroup1:subgroup2:role=admin#authority",
            "authority",
        ),
        ("urn:nid:ns1:ns2:ns3:group:group:subgroup1:subgroup2:role=admin#.", "."),
        (
            "urn:geant:h-df.de:group:IMK-TRO-EWCC#login.helmholtz-data-federation.de",
            "login.helmholtz-data-federation.de",
        ),
    ],
)
def test_entitlement_group_authority(raw, group_authority):
    ent = Entitlement(raw)
    assert ent.group_authority == group_authority


@pytest.mark.parametrize(
    "raw",
    [
        "urn:nid:namespace:group:group:subgroup1:subgroup2:role=admin#authority",  # all fields
        "urn:nid:namespace:group:group:subgroup1:subgroup2#authority",  # no role
        "urn:nid:namespace:group:group:role=admin#authority",  # no subgroups
        "urn:geant:h-df.de:group:IMK-TRO-EWCC#login.helmholtz-data-federation.de",
        "urn:geant:h-df.de:group:MyExampleColab#login.helmholtz-data-federation.de",
        "urn:geant:h-df.de:group:wlcg-test#login.helmholtz-data-federation.de",
        "urn:geant:h-df.de:group:HDF#login.helmholtz-data-federation.de",
    ],
)
def test_entitlement_valid(raw):
    ent = Entitlement(raw)
    assert repr(ent) == raw


@pytest.mark.parametrize(
    "raw",
    [
        "nid:namespace:group:group:subgroup1:subgroup2:role=admin#authority",
        "urn:nid:namespace:group::subgroup1:subgroup2:role=admin#authority",
        "urn:nid:namespace:group:group:subgroup1:subgroup2:role=admin#",
        "urn:group:group:subgroup1:subgroup2:role=admin#authority",
        "urn:nid:namespace:subgroup1:subgroup2:role=admin#authority",
        "urn:nid:namespace:group:group:subgroup1:subgroup2:role=admin",
        "urn:nid:namespace:group:group:subgroup1:subgroup2",
        "urn:nid:namespace:group:group:subgroup1",
        "urn:nid:namespace:group:group",
        "urn:nid:namespace:group",
        "urn:nid:namespace",
        "urn:nid",
        "anything",
    ],
)
def test_entitlement_invalid(raw):
    with pytest.raises(ValueError):
        Entitlement(raw)



#!/usr/bin/env python3
# pylint
# vim: tw=100
# pylint: disable=bad-continuation, invalid-name, superfluous-parens
# pylint: disable=bad-whitespace, mixed-indentation
# pylint: disable=redefined-outer-name, logging-not-lazy, logging-format-interpolation
# pylint: disable=missing-docstring, trailing-whitespace, trailing-newlines, too-few-public-methods

import sys
import os
import re
import json
import requests
import logging
import argparse
from pathlib import Path
from configparser import ConfigParser
from configparser import ExtendedInterpolation
import urllib.parse as ul

# CONFIG = ConfigParser()
CONFIG = ConfigParser(interpolation=ExtendedInterpolation())
CONFIG.optionxform = lambda option: option

# Logging
logformat = "[%(levelname)s] [%(filename)s:%(funcName)s:%(lineno)d] %(message)s"
logging.basicConfig(level=os.environ.get("LOG", "WARN"), format=logformat)
logger = logging.getLogger(__name__)


# Functions
def load_config():
    """Reload configuration from disk.

    Config locations, by priority (all values are merged. The last one
        overwrites earlier ones)
    """
    files = []
    try:
        files += [Path(args.pathconf_file)]
    except:
        pass

    logger.info("reading config")

    files += [
        Path("./subiss-to-unix.conf"),
        Path(Path.home(), ".config", "subiss-to-unix.conf"),
        Path("/etc/subiss-to-unix.conf"),
    ]

    for f in files:
        if f.exists():
            logger.info("Using this config file: {}".format(f))
            CONFIG.read(f)
            break


def remove_quotes(data):
    return data.lstrip('"').lstrip("'").rstrip('"').rstrip("'")


def parseOptions():
    """Parse the commandline options"""

    path_of_executable = os.path.realpath(sys.argv[0])
    folder_of_executable = os.path.split(path_of_executable)[0]
    full_name_of_executable = os.path.split(path_of_executable)[1]
    name_of_executable = full_name_of_executable.rstrip(".py")

    parser = argparse.ArgumentParser()

    parser.add_argument("--rest_user", "-u", help="username for LDF rest interface")
    parser.add_argument("--rest_passwd", "-p", help="passwdname for LDF rest interface")
    parser.add_argument("--iss", action="append")
    parser.add_argument("--bwidmOrgId", default="hdf")
    parser.add_argument("--base_url", default="https://bwidm-test.scc.kit.edu/rest/")
    parser.add_argument(
        "--verify_tls", default=True, action="store_false", help="disable verify"
    )
    parser.add_argument(
        "--issTranslateExpression",
        default='{"unity-hdf": "unity.helmholtz-data-federation.de/oauth2",  "kit": "https://oidc.scc.kit.edu/auth/realms/kit"}',
    )
    parser.add_argument(
        dest="sub_iss",
        help='Content of $REMOTE_USER. For testing use "test-offline" and "test-id"',
    )
    parser.add_argument("--verbose", "-v", default=False, action="store_true")
    parser.add_argument("--debug", "-d", default=False, action="store_true")
    args = parser.parse_args()

    # sanitize some args:
    args.base_url = args.base_url.rstrip("/")
    args.bwidmOrgId = remove_quotes(args.bwidmOrgId)

    # ensure translation will work as JSON
    try:
        args.issTranslateExpressionJSON = json.loads(args.issTranslateExpression)
        try:
            for key in args.issTranslateExpressionJSON.keys():
                args.issTranslateExpressionJSON[key] = remove_quotes(
                    args.issTranslateExpressionJSON[key]
                )
        except:
            sys.stderr.write(
                "FATAL: issTranslateExpression needs to be a one line json object that lists keys and values: \n"
            )
            sys.stderr.write(
                '{"unity-hdf": "unity.helmholtz-data-federation.de/oauth2", "test": "https://test.com"}'
            )
            sys.stderr.write(
                'Instead you provided "%s"\n' % str(args.issTranslateExpression)
            )
            raise
    except:
        raise

    return args


args = parseOptions()
load_config()

if args.sub_iss == "test-offline":
    sys.stdout.write("hdf_marcus\n")
    exit(0)

if args.sub_iss == "test-id":
    args.sub_iss = "6c611e2a-2c1c-487f-9948-c058a36c8f0e@https://login.helmholtz-data-federation.de/oauth2"
    sys.stderr.write("using test id: %s\n" % args.sub_iss)
if args.sub_iss == "test-marcus":
    args.sub_iss = "6c611e2a-2c1c-487f-9948-c058a36c8f0e@https://login.helmholtz-data-federation.de/oauth2"
    sys.stderr.write("using test id: %s\n" % args.sub_iss)
if args.sub_iss == "test-borja-old":
    args.sub_iss = "d9f4d895-6051-4717-883e-4b2676ad0d0d@https://login.helmholtz-data-federation.de/oauth2"
    sys.stderr.write("using test id: %s\n" % args.sub_iss)
if args.sub_iss == "test-borja-new":
    args.sub_iss = "309ed509-c56a-4894-b163-5993bd08cbc2@https://login.helmholtz-data-federation.de/oauth2"
    sys.stderr.write("using test id: %s\n" % args.sub_iss)

externalId = args.sub_iss
vals = args.sub_iss.split("@")
sub = "@".join(vals[0:-1])  # First few components are sub, may contain '@'
iss = vals[-1]  # Last component is issuer may NOT contain '@'
externalId = ul.quote_plus(sub) + "@n" + ul.quote_plus(iss)

url = args.base_url + "/external-user/find/externalId/" + ul.quote_plus(str(externalId))
rest_user = CONFIG["main"].get("rest_user", "xxx")
rest_pass = CONFIG["main"].get("rest_pass", "xxx")
base_url = CONFIG["main"].get("base_url", "")

if args.verbose:
    logger.debug(f"URL: {url}")

resp = requests.get(url, verify=args.verify_tls, auth=(rest_user, rest_pass))

if resp.status_code != 200:
    sys.stderr.write(
        "Error %d reading from remote: \n%s\n" % (resp.status_code, str(resp.text))
    )
    exit(1)

resp_json = resp.json()
try:
    username = resp_json["attributeStore"]["urn:oid:0.9.2342.19200300.100.1.1"]
    bwIdmOrgId = resp_json["attributeStore"]["http://bwidm.de/bwidmOrgId"]

    sys.stdout.write("%s_%s\n" % (bwIdmOrgId, username))
except KeyError as e:
    if args.verbose:
        # sys.stderr.write('Error interpreting remote json object: Key Error: %s not found\n' % str(e))
        sys.stderr.write(
            "Error: I could not find the username in the database. Most likely the user is not registered for this service\n"
        )
        sys.stderr.write("This is the json data received\n")
        sys.stderr.write(
            json.dumps(resp_json, sort_keys=True, indent=4, separators=(",", ": "))
        )
        sys.stderr.write("\n")

if args.debug:
    sys.stderr.write("This is the json data received\n")
    sys.stderr.write(
        json.dumps(resp_json, sort_keys=True, indent=4, separators=(",", ": "))
    )
    sys.stderr.write("\n")



#!/usr/bin/env python3
# vim: tw=100 foldmethod=indent
# pylint: disable=invalid-name, superfluous-parens
# pylint: disable=redefined-outer-name, logging-not-lazy, logging-format-interpolation, logging-fstring-interpolation
# pylint: disable=missing-docstring, trailing-whitespace, trailing-newlines, too-few-public-methods
#
# Author: Joshua Bachmeier <joshua.bachmeier@student.kit.edu>
#

import sys
import json
import logging
from ldf_adapter import logsetup
from ldf_adapter.logsetup import jsonlogger

# Must be before the first ldf_adapter import
from feudal_globalconfig import globalconfig

from ldf_adapter import User
from ldf_adapter.results import ExceptionalResult, FatalError
from ldf_adapter.cmdline_params import args

logger = logging.getLogger(__name__)


class PathTruncatingFormatter(logging.Formatter):
    """formatter for logging"""

    def format(self, record):
        pathname = record.pathname
        if len(pathname) > 23:
            pathname = "...{}".format(pathname[-19:])
        record.pathname = pathname
        return super(PathTruncatingFormatter, self).format(record)


def main():
    if args.test:
        logger.info("test mode")
        data = {
            "state_target": "test",
            "user": {
                "userinfo": {
                    "sub": "test",
                    "iss": "test",
                }
            },
        }
    else:
        try:
            data = json.load(sys.stdin)
        except json.decoder.JSONDecodeError as e:
            message = "Cannot decode the input json. Please verify the input!"
            logger.error(message)
            raise FatalError(message=message)

    logger.debug(f"Attempting to reach state '{data['state_target']}'")

    if data["user"]["userinfo"] is None:
        message = "Cannot process null input"
        logger.error(message)
        raise FatalError(message=message)

    try:
        result = User(data).reach_state(data["state_target"])
    except ExceptionalResult as result:
        result = result.attributes
        logger.debug("Reached state '{state}': {message}".format(**result))
        json.dump(result, sys.stdout)
    else:
        result = result.attributes
        logger.debug("Reached state '{state}': {message}".format(**result))
        json.dump(result, sys.stdout)

    return 0


if __name__ == "__main__":
    sys.exit(main())



class MockUserInfo:
    """Mocks a UserInfo object to be passed to backend user
    Only a few properties are necessary:
        - unique_id
        - username
        - primary_group
        - ssh_keys
    """

    def __init__(self, data):
        self.unique_id = data.get("unique_id")
        self.username = data.get("username")
        self.primary_group = data.get("primary_group")
        self.ssh_keys = data.get("ssh_keys")
        self.family_name = data.get("family_name")
        self.given_name = data.get("given_name")
        self.full_name = data.get("full_name")
        self.email = data.get("email")


def wrap_userinfo(input):
    return {"user": {"userinfo": input}}


INPUT_UNITY = wrap_userinfo(
    {
        "display_name": "Marcus Hardt",
        "eduperson_assurance": [
            "https://refeds.org/assurance/IAP/medium",
            "https://refeds.org/assurance/IAP/local-enterprise",
            "https://refeds.org/assurance/ID/eppn-unique-no-reassign",
            "https://refeds.org/assurance/ATP/ePA-1m",
            "https://refeds.org/assurance/ATP/ePA-1d",
            "https://refeds.org/assurance/ID/unique",
            "https://refeds.org/assurance/profile/cappuccino",
            "https://refeds.org/assurance/IAP/low",
        ],
        "eduperson_entitlement": [
            "urn:geant:h-df.de:group:IMK-TRO-EWCC#login.helmholtz-data-federation.de",
            "urn:geant:h-df.de:group:MyExampleColab#login.helmholtz-data-federation.de",
            "urn:geant:h-df.de:group:wlcg-test#login.helmholtz-data-federation.de",
            "urn:geant:h-df.de:group:HDF#login.helmholtz-data-federation.de",
        ],
        "eduperson_principal_name": "lo0018@kit.edu",
        "eduperson_scoped_affiliation": "member@kit.edu",
        "eduperson_unique_id": "6c611e2a2c1c487f9948c058a36c8f0e@login.helmholtz-data-federation.de",
        "email": "marcus.hardt@kit.edu",
        "email_verified": "true",
        "family_name": "Hardt",
        "given_name": "Marcus",
        "groups": ["/wlcg-test", "/IMK-TRO-EWCC", "/MyExampleColab", "/HDF", "/"],
        "name": "Marcus Hardt",
        "preferred_username": "marcus",
        "sn": "Hardt",
        "ssh_key": "ssh-ed25519 AAAAC3NzaC1lZDI1NTE5AAAAIAqA5FW6m3FbFhCOsRQBxKMRki5qJxoNhZdaeLXg6ym/ marcus@nemo2019\n",
        "sub": "6c611e2a-2c1c-487f-9948-c058a36c8f0e",
        "iss": "https://login.helmholtz-data-federation.de/oauth2",
    }
)

INPUT_EGI = wrap_userinfo(
    {
        "acr": "https://aai.egi.eu/LoA#Substantial",
        "eduperson_assurance": ["https://aai.egi.eu/LoA#Substantial"],
        "email": "marcus.hardt@kit.edu",
        "family_name": "Hardt",
        "given_name": "Marcus",
        "preferred_username": "mhardt",
        "sub": "d7a53cbe3e966c53ac64fde7355956560282158ecac8f3d2c770b474862f4756@egi.eu",
        "iss": "https://aai.egi.eu/oidc/",
    }
)
INPUT_EGI_MANYGROUPS = wrap_userinfo(
    {
        "eduperson_assurance": [
            "https://refeds.org/assurance/ATP/ePA-1d",
            "https://refeds.org/assurance/ATP/ePA-1m",
            "https://refeds.org/assurance/IAP/low",
            "https://refeds.org/assurance/profile/cappuccino",
            "https://refeds.org/assurance/IAP/local-enterprise",
            "https://refeds.org/assurance/IAP/medium",
            "https://refeds.org/assurance/ID/unique",
            "https://refeds.org/assurance/ID/eppn-unique-no-reassign",
            "https://aai.egi.eu/LoA#Substantial",
        ],
        "eduperson_entitlement": [
            "urn:mace:egi.eu:group:covid19.eosc-synergy.eu:admins:role=member#aai.egi.eu",
            "urn:mace:egi.eu:group:covid19.eosc-synergy.eu:admins:role=owner#aai.egi.eu",
            "urn:mace:egi.eu:group:cryoem.instruct-eric.eu:admins:role=member#aai.egi.eu",
            "urn:mace:egi.eu:group:cryoem.instruct-eric.eu:admins:role=owner#aai.egi.eu",
            "urn:mace:egi.eu:group:eosc-synergy.eu:admins:role=member#aai.egi.eu",
            "urn:mace:egi.eu:group:eosc-synergy.eu:admins:role=owner#aai.egi.eu",
            "urn:mace:egi.eu:group:eosc-synergy.eu:role=member#aai.egi.eu",
            "urn:mace:egi.eu:group:eosc-synergy.eu:role=vm_operator#aai.egi.eu",
            "urn:mace:egi.eu:group:goc.egi.eu:role=member#aai.egi.eu",
            "urn:mace:egi.eu:group:goc.egi.eu:role=vm_operator#aai.egi.eu",
            "urn:mace:egi.eu:group:mteam.data.kit.edu:admins:role=member#aai.egi.eu",
            "urn:mace:egi.eu:group:mteam.data.kit.edu:admins:role=owner#aai.egi.eu",
            "urn:mace:egi.eu:group:mteam.data.kit.edu:perfmon.m.d.k.e:admins:role=member#aai.egi.eu",
            "urn:mace:egi.eu:group:mteam.data.kit.edu:perfmon.m.d.k.e:role=member#aai.egi.eu",
            "urn:mace:egi.eu:group:mteam.data.kit.edu:role=member#aai.egi.eu",
            "urn:mace:egi.eu:group:mteam.data.kit.edu:role=vm_operator#aai.egi.eu",
            "urn:mace:egi.eu:group:o3as.data.kit.edu:admins:role=member#aai.egi.eu",
            "urn:mace:egi.eu:group:o3as.data.kit.edu:admins:role=owner#aai.egi.eu",
            "urn:mace:egi.eu:group:o3as.data.kit.edu:role=member#aai.egi.eu",
            "urn:mace:egi.eu:group:o3as.data.kit.edu:role=vm_operator#aai.egi.eu",
            "urn:mace:egi.eu:group:registry:perfmon:role=member#aai.egi.eu",
            "urn:mace:egi.eu:group:registry:perfmon:role=owner#aai.egi.eu",
            "urn:mace:egi.eu:group:saps-vo.i3m.upv.es:admins:role=member#aai.egi.eu",
            "urn:mace:egi.eu:group:saps-vo.i3m.upv.es:admins:role=owner#aai.egi.eu",
            "urn:mace:egi.eu:group:umsa.cerit-sc.cz:admins:role=member#aai.egi.eu",
            "urn:mace:egi.eu:group:umsa.cerit-sc.cz:admins:role=owner#aai.egi.eu"
            "urn:mace:egi.eu:group:university.eosc-synergy.eu:admins:role=member#aai.egi.eu",
            "urn:mace:egi.eu:group:university.eosc-synergy.eu:admins:role=owner#aai.egi.eu",
            "urn:mace:egi.eu:group:university.eosc-synergy.eu:role=member#aai.egi.eu",
            "urn:mace:egi.eu:group:university.eosc-synergy.eu:role=vm_operator#aai.egi.eu",
            "urn:mace:egi.eu:group:worsica.vo.incd.pt:role=member#aai.egi.eu",
            "urn:mace:egi.eu:group:worsica.vo.incd.pt:role=vm_operator#aai.egi.eu",
        ],
        "eduperson_scoped_affiliation": ["employee@kit.edu", "member@kit.edu"],
        "eduperson_unique_id": "d7a53cbe3e966c53ac64fde7355956560282158ecac8f3d2c770b474862f4756@egi.eu",
        "email": "hardt@kit.edu",
        "email_verified": True,
        "family_name": "Hardt",
        "given_name": "Marcus",
        "iss": "https://aai.egi.eu/auth/realms/egi",
        "name": "Marcus Hardt",
        "preferred_username": "mhardt",
        "sub": "d7a53cbe3e966c53ac64fde7355956560282158ecac8f3d2c770b474862f4756@egi.eu",
        "voperson_verified_email": ["hardt@kit.edu"],
    }
)


INPUT_DEEP_IAM = wrap_userinfo(
    {
        "iss": "https://iam.deep-hybrid-datacloud.eu/",
        "family_name": "Hardt",
        "given_name": "Marcus",
        "groups": ["KIT-Cloud"],
        "name": "Marcus Hardt",
        "organisation_name": "deep-hdc",
        "preferred_username": "marcus",
        "sub": "d9730f60-3b19-4f45-83ab-f29addf72d58",
        "updated_at": "Mon Jun 25 16:55:15 CEST 2018",
    }
)

INPUT_INDIGO_IAM = wrap_userinfo(
    {
        "iss": "https://iam-test.indigo-datacloud.eu/",
        "family_name": "Hardt",
        "gender": "M",
        "given_name": "Marcus",
        "groups": ["Users", "Developers", "test.vo-users"],
        "name": "Marcus Hardt",
        "organisation_name": "indigo-dc",
        "preferred_username": "marcus",
        "sub": "a1ea3aa2-8daf-41bb-b4fb-eb88f439e446",
        "updated_at": 1563283972,
    }
)

INPUT_KIT = wrap_userinfo(
    {
        "displayName": "Hardt, Marcus (SCC)",
        "eduPersonEntitlement": [
            "urn:geant:kit.edu:group:DFN-SLCS#kit.edu",
            "urn:geant:kit.edu:group:LSDF-DIS#kit.edu",
            "urn:geant:kit.edu:group:bwGrid#kit.edu",
            "urn:geant:kit.edu:group:bwLSDF-FS#kit.edu",
            "urn:geant:kit.edu:group:bwUniCluster#kit.edu",
            "urn:geant:kit.edu:group:bwsyncnshare#kit.edu",
            "urn:geant:kit.edu:group:bwsyncnshare-idm#kit.edu",
            "urn:geant:kit.edu:group:gruppenverwalter#kit.edu",
        ],
        "eduPersonPrincipalName": "lo0018@kit.edu",
        "eduPersonScopedAffiliation": ["employee@kit.edu", "member@kit.edu"],
        "eduperson_entitlement": [
            "urn:geant:kit.edu:group:DFN-SLCS#kit.edu",
            "urn:geant:kit.edu:group:LSDF-DIS#kit.edu",
            "urn:geant:kit.edu:group:bwGrid#kit.edu",
            "urn:geant:kit.edu:group:bwLSDF-FS#kit.edu",
            "urn:geant:kit.edu:group:bwUniCluster#kit.edu",
            "urn:geant:kit.edu:group:bwsyncnshare#kit.edu",
            "urn:geant:kit.edu:group:bwsyncnshare-idm#kit.edu",
            "urn:geant:kit.edu:group:gruppenverwalter#kit.edu",
        ],
        "eduperson_principal_name": "lo0018@kit.edu",
        "eduperson_scoped_affiliation": ["employee@kit.edu", "member@kit.edu"],
        "email": "marcus.hardt@kit.edu",
        "family_name": "Hardt",
        "givenName": "Marcus",
        "given_name": "Marcus",
        "mail": "marcus.hardt@kit.edu",
        "name": "Marcus Hardt",
        "ou": "SCC",
        "preferred_username": "lo0018",
        "sn": "Hardt",
        "sub": "4cbcd471-1f51-4e54-97b8-2dd5177e25ec",
        "iss": "https://oidc.scc.kit.edu/auth/realms/kit/",
    }
)


ALL_INPUT = [INPUT_UNITY, INPUT_EGI, INPUT_DEEP_IAM, INPUT_INDIGO_IAM, INPUT_KIT]



# vim: foldmethod=indent : tw=100
import pytest
import logging
import ldap3
import sys

from ..settings import MockUserInfo
from ldf_adapter.results import Rejection, Failure


logger = logging.getLogger(__name__)


original_init = ldap3.Connection.__init__


def mock_ldap3_connection_init(self, *args, **kwargs):
    if kwargs.get("client_strategy") in [ldap3.SAFE_SYNC, ldap3.SAFE_RESTARTABLE]:
        kwargs["client_strategy"] = ldap3.MOCK_SYNC
    if kwargs.get("auto_bind"):
        del kwargs["auto_bind"]
    original_init(self, *args, **kwargs)
    self.bind()


@pytest.fixture(scope="function")
def ldap_user(mode, userinfo, existing_entries, monkeypatch):
    """Creates a backend user from provided dict data.
    'userinfo' should contain: unique_id, username, primary_group.
    'existing_entries' are LDAP entries to be added beforehand;
        each entry contains a dn and a dict of attributes.
    """
    with monkeypatch.context() as mp:
        mp.setattr("ldap3.Connection.__init__", mock_ldap3_connection_init)
        mp.setattr("ldf_adapter.backend.ldap.CONFIG.backend.ldap.mode", mode)

        import ldf_adapter.backend.ldap

        service_user = ldf_adapter.backend.ldap.User(MockUserInfo(userinfo))
        for entry in existing_entries:
            ldf_adapter.backend.ldap.LDAP.connection.add(
                entry["dn"],
                object_class=entry["object_class"],
                attributes=entry["attributes"],
            )
        yield service_user
        # module needs to be reloaded on next test so that LDAP object is recreated
        del sys.modules["ldf_adapter.backend.ldap"]


@pytest.fixture(scope="function")
def ldap_group(mode, name, existing_entries, monkeypatch):
    """Creates a backend group from provided data.
    name: the group name (no constraints on allowed names)
    'existing_entries' are LDAP entries to be added beforehand;
        each entry contains a dn and a dict of attributes.
    """
    with monkeypatch.context() as mp:
        mp.setattr("ldap3.Connection.__init__", mock_ldap3_connection_init)
        mp.setattr("ldf_adapter.backend.ldap.CONFIG.backend.ldap.mode", mode)

        import ldf_adapter.backend.ldap

        service_group = ldf_adapter.backend.ldap.Group(name)
        for entry in existing_entries:
            ldf_adapter.backend.ldap.LDAP.connection.add(
                entry["dn"],
                object_class=entry["object_class"],
                attributes=entry["attributes"],
            )
        yield service_group
        del sys.modules["ldf_adapter.backend.ldap"]


@pytest.fixture(scope="function")
def ldap_mod_entry(
    mode, userinfo, supplementary_group_names, existing_entries, monkeypatch
):
    """Creates a backend user and groups from provided data.
    'userinfo' should contain: unique_id, username, primary_group.
    'supplementary_group_names' is a list of group names
    'existing_entries' are LDAP entries to be added beforehand;
        each entry contains a dn and a dict of attributes.
    """
    with monkeypatch.context() as mp:
        mp.setattr("ldap3.Connection.__init__", mock_ldap3_connection_init)
        mp.setattr("ldf_adapter.backend.ldap.CONFIG.backend.ldap.mode", mode)

        import ldf_adapter.backend.ldap

        service_user = ldf_adapter.backend.ldap.User(MockUserInfo(userinfo))
        supplementary_groups = []
        for name in supplementary_group_names:
            supplementary_groups.append(ldf_adapter.backend.ldap.Group(name))
        for entry in existing_entries:
            ldf_adapter.backend.ldap.LDAP.connection.add(
                entry["dn"],
                object_class=entry["object_class"],
                attributes=entry["attributes"],
            )
        yield {"user": service_user, "supplementary_groups": supplementary_groups}
        del sys.modules["ldf_adapter.backend.ldap"]


@pytest.fixture(scope="function")
def ldap_connection(uid_min, uid_max, gid_min, gid_max, monkeypatch):
    with monkeypatch.context() as mp:
        mp.setattr("ldap3.Connection.__init__", mock_ldap3_connection_init)
        mp.setattr("ldf_adapter.backend.ldap.CONFIG.backend.ldap.mode", "full_access")
        mp.setattr("ldf_adapter.backend.ldap.CONFIG.backend.ldap.uid_min", uid_min)
        mp.setattr("ldf_adapter.backend.ldap.CONFIG.backend.ldap.uid_max", uid_max)
        mp.setattr("ldf_adapter.backend.ldap.CONFIG.backend.ldap.gid_min", gid_min)
        mp.setattr("ldf_adapter.backend.ldap.CONFIG.backend.ldap.gid_max", gid_max)
        import ldf_adapter.backend.ldap

        yield ldf_adapter.backend.ldap.LDAP
        del sys.modules["ldf_adapter.backend.ldap"]


USERINFO = {
    "unique_id": "subuid@issuer.domain",
    "primary_group": "testgroup",
    "username": "testuser",
    "ssh_keys": {},
}

USERINFO_BIG = {
    "unique_id": "subuid@issuer.domain",
    "primary_group": "testgroup",
    "username": "testuser",
    "ssh_keys": {},
    "family_name": "family_name",
    "given_name": "given_name",
    "full_name": "full_name",
    "email": "email@domain",
}

LDAP_USER_ENTRY = {
    "dn": "uid=testuser,ou=users,dc=example",
    "object_class": ["top", "inetOrgPerson", "posixAccount"],
    "attributes": {
        "uid": "testuser",
        "uidNumber": 1000,
        "gidNumber": 1000,
        "homeDirectory": "/home/testuser",
        "loginShell": "/bin/sh",
        "gecos": "subuid@issuer.domain",
        "username": "testuser",
    },
}

LDAP_USER2_ENTRY = {
    "dn": "uid=testuser,ou=users,dc=example",
    "object_class": ["top", "inetOrgPerson", "posixAccount"],
    "attributes": {
        "uid": "testuser",
        "uidNumber": 1000,
        "gidNumber": 1000,
        "homeDirectory": "/home/testuser",
        "loginShell": "/bin/sh",
        "gecos": "another_uid@issuer.domain",
        "username": "testuser",
    },
}

LDAP_TESTGROUP_ENTRY = {
    "dn": "cn=testgroup,ou=groups,dc=example",
    "object_class": ["posixGroup"],
    "attributes": {"cn": "testgroup", "gidNumber": 1000, "memberUid": ["testuser"]},
}

LDAP_TESTGROUP2_ENTRY = {
    "dn": "cn=testgroup2,ou=groups,dc=example",
    "object_class": ["posixGroup"],
    "attributes": {"cn": "testgroup2", "gidNumber": 1001, "memberUid": ["testuser"]},
}

LDAP_GROUP1_ENTRY = {
    "dn": "cn=group1,ou=groups,dc=example",
    "object_class": ["posixGroup"],
    "attributes": {
        "cn": "group1",
        "gidNumber": 1002,
    },
}

LDAP_GROUP2_ENTRY = {
    "dn": "cn=group2,ou=groups,dc=example",
    "object_class": ["posixGroup"],
    "attributes": {
        "cn": "group2",
        "gidNumber": 1003,
    },
}

LDAP_PRECREATED_USER_ENTRY = {
    "dn": "uid=testuser,ou=users,dc=example",
    "object_class": ["top", "inetOrgPerson", "posixAccount"],
    "attributes": {
        "uid": "testuser",
        "uidNumber": 1000,
        "gidNumber": 1000,
        "homeDirectory": "/home/testuser",
        "loginShell": "/bin/sh",
        "username": "testuser",
    },
}

LDAP_PRECREATED_TESTGROUP_ENTRY = {
    "dn": "cn=testgroup,ou=groups,dc=example",
    "object_class": ["posixGroup"],
    "attributes": {
        "cn": "testgroup",
        "gidNumber": 1000,
    },
}


@pytest.mark.parametrize("userinfo", [USERINFO])
@pytest.mark.parametrize("mode", ["read_only"])
@pytest.mark.parametrize("existing_entries", [[], [LDAP_USER_ENTRY]])
def test_create_user_read_only_fail(ldap_user, existing_entries):
    """In read_only mode, create always fails, even if user exists."""
    assert ldap_user.exists() == (len(existing_entries) > 0)
    with pytest.raises(Rejection):
        ldap_user.create()


@pytest.mark.parametrize(
    "mode,userinfo,existing_entries",
    [
        ("pre_created", USERINFO, []),
    ],
)
def test_create_user_precreated_doesnt_exist(ldap_user):
    """In pre_created mode, create fails if entry for username does not exist."""
    assert not ldap_user.exists()
    with pytest.raises(Rejection):
        ldap_user.create()


@pytest.mark.parametrize(
    "mode,userinfo,existing_entries",
    [
        (
            "pre_created",
            USERINFO,
            [LDAP_PRECREATED_USER_ENTRY, LDAP_PRECREATED_TESTGROUP_ENTRY],
        ),
    ],
)
def test_create_user_precreated_exists(ldap_user):
    """In pre_created mode, create succeeds if entry for username exists by mapping entry to unique_id.
    Will not check if entry already mapped to something else, since ldf_adapter user checks if
    user exists or if name_taken beforehand.
    """
    assert not ldap_user.exists()
    ldap_user.create()
    assert ldap_user.exists()


@pytest.mark.parametrize(
    "mode,userinfo,existing_entries",
    [
        ("full_access", USERINFO, []),
    ],
)
def test_create_user_full_access_group_doesnt_exist(ldap_user):
    """primary group must exist, otherwise create fails."""
    assert not ldap_user.exists()
    with pytest.raises(Failure):
        ldap_user.create()


@pytest.mark.parametrize(
    "mode,userinfo,existing_entries",
    [
        ("full_access", USERINFO, [LDAP_PRECREATED_TESTGROUP_ENTRY]),
    ],
)
def test_create_user_full_access_success(ldap_user):
    assert not ldap_user.exists()
    ldap_user.create()
    assert ldap_user.exists()


@pytest.mark.parametrize("userinfo", [USERINFO])
@pytest.mark.parametrize("mode", ["read_only", "pre_created", "full_access"])
@pytest.mark.parametrize(
    "existing_entries,username",
    [
        ([], None),
        ([LDAP_USER_ENTRY], "testuser"),
    ],
)
def test_create_user_get_username(ldap_user, username):
    assert ldap_user.get_username() == username


@pytest.mark.parametrize("userinfo", [USERINFO])
@pytest.mark.parametrize("mode", ["read_only", "pre_created", "full_access"])
@pytest.mark.parametrize(
    "existing_entries,groupname",
    [
        ([], None),
        ([LDAP_USER_ENTRY], None),
        ([LDAP_USER_ENTRY, LDAP_PRECREATED_TESTGROUP_ENTRY], "testgroup"),
        ([LDAP_USER_ENTRY, LDAP_TESTGROUP_ENTRY], "testgroup"),
    ],
)
def test_get_primary_group(ldap_user, groupname):
    assert ldap_user.get_primary_group() == groupname


@pytest.mark.parametrize("userinfo", [USERINFO])
@pytest.mark.parametrize("mode", ["read_only", "pre_created", "full_access"])
@pytest.mark.parametrize("existing_entries", [[]])
def test_name_taken_no_existing_entries(ldap_user):
    assert not ldap_user.name_taken("testuser")


@pytest.mark.parametrize("userinfo", [USERINFO])
@pytest.mark.parametrize("mode", ["read_only", "pre_created", "full_access"])
@pytest.mark.parametrize("existing_entries", [[LDAP_USER_ENTRY]])
def test_name_taken_by_user(ldap_user):
    assert not ldap_user.name_taken("testuser")


@pytest.mark.parametrize("userinfo", [USERINFO])
@pytest.mark.parametrize("mode", ["read_only", "pre_created", "full_access"])
@pytest.mark.parametrize("existing_entries", [[LDAP_USER2_ENTRY]])
def test_name_taken_by_another(ldap_user):
    assert ldap_user.name_taken("testuser")


@pytest.mark.parametrize(
    "mode,userinfo,existing_entries",
    [
        ("pre_created", USERINFO, [LDAP_PRECREATED_USER_ENTRY]),
    ],
)
def test_name_taken_precreated_not_mapped(ldap_user):
    assert not ldap_user.name_taken("testuser")


@pytest.mark.parametrize("userinfo", [USERINFO])
@pytest.mark.parametrize("mode", ["read_only"])
@pytest.mark.parametrize(
    "existing_entries,groups",
    [
        ([], []),
        ([LDAP_USER_ENTRY], []),
        ([LDAP_USER_ENTRY, LDAP_PRECREATED_TESTGROUP_ENTRY], []),
        ([LDAP_USER_ENTRY, LDAP_TESTGROUP_ENTRY], ["testgroup"]),
        (
            [LDAP_USER_ENTRY, LDAP_TESTGROUP_ENTRY, LDAP_TESTGROUP2_ENTRY],
            ["testgroup", "testgroup2"],
        ),
    ],
)
def test_get_groups(ldap_user, groups):
    assert set(ldap_user.get_groups()) == set(groups)


@pytest.mark.parametrize(
    "mode,userinfo,existing_entries",
    [
        ("full_access", USERINFO, []),
        ("full_access", USERINFO, [LDAP_USER_ENTRY]),
    ],
)
def test_delete_user_sucecss(ldap_user, existing_entries):
    """works no matter if user exists or not"""
    assert ldap_user.exists() == bool(len(existing_entries))
    ldap_user.delete()
    assert not ldap_user.exists()


@pytest.mark.parametrize("mode", ["read_only", "pre_created"])
@pytest.mark.parametrize(
    "userinfo,existing_entries",
    [
        (USERINFO, [LDAP_USER_ENTRY]),
        (USERINFO, []),
    ],
)
def test_delete_user_rejected(ldap_user):
    """rejected no matter if user exists or not"""
    with pytest.raises(Rejection):
        ldap_user.delete()


@pytest.mark.parametrize("mode", ["read_only", "pre_created"])
@pytest.mark.parametrize(
    "userinfo,existing_entries",
    [
        (USERINFO, [LDAP_USER_ENTRY]),
        (USERINFO, []),
    ],
)
def test_update_user_ignored(ldap_user, existing_entries):
    """in read-only and per-created mode, nothing happens."""
    assert ldap_user.exists() == bool(len(existing_entries))
    ldap_user.update()
    assert ldap_user.exists() == bool(len(existing_entries))


@pytest.mark.parametrize(
    "mode,userinfo,existing_entries",
    [
        ("full_access", USERINFO, []),
    ],
)
def test_update_full_access_doesnt_exist(ldap_user):
    assert not ldap_user.exists()
    ldap_user.update()
    assert not ldap_user.exists()


@pytest.mark.parametrize(
    "mode,userinfo,existing_entries",
    [
        ("full_access", USERINFO, [LDAP_USER_ENTRY]),
        ("full_access", USERINFO_BIG, [LDAP_USER_ENTRY]),
    ],
)
def test_update_full_access_new_attribute(ldap_user, userinfo):
    assert ldap_user.exists()
    ldap_user.update()
    assert ldap_user.exists()
    attributes = ldap_user.get_ldap_entry()
    assert attributes.get("sn") == userinfo.get("family_name")
    assert attributes.get("givenName") == userinfo.get("given_name")
    assert attributes.get("cn") == userinfo.get("full_name")
    assert attributes.get("mail") == userinfo.get("email")


@pytest.mark.parametrize("mode", ["read_only", "pre_created"])
@pytest.mark.parametrize("name", ["testgroup"])
@pytest.mark.parametrize(
    "existing_entries",
    [
        [LDAP_TESTGROUP_ENTRY],
        [],
    ],
)
def test_create_group_ignore(ldap_group, existing_entries):
    assert ldap_group.exists() == bool(len(existing_entries))
    ldap_group.create()
    assert ldap_group.exists() == bool(len(existing_entries))


@pytest.mark.parametrize("mode", ["full_access"])
@pytest.mark.parametrize("name", ["testgroup"])
@pytest.mark.parametrize(
    "existing_entries",
    [
        [LDAP_TESTGROUP_ENTRY],
        [],
    ],
)
def test_create_group_success(ldap_group, existing_entries):
    assert ldap_group.exists() == bool(len(existing_entries))
    ldap_group.create()
    assert ldap_group.exists()


@pytest.mark.parametrize("mode", ["read_only", "pre_created"])
@pytest.mark.parametrize("userinfo", [USERINFO])
@pytest.mark.parametrize(
    "existing_entries",
    [
        [LDAP_USER_ENTRY],
        [],
    ],
)
@pytest.mark.parametrize("supplementary_group_names", [[], ["group1", "group2"]])
def test_mod_user_ignore(ldap_mod_entry):
    """nothing happens"""
    added, removed = ldap_mod_entry["user"].mod(ldap_mod_entry["supplementary_groups"])
    assert added == []
    assert removed == []


@pytest.mark.parametrize("mode", ["pre_created", "full_access"])
@pytest.mark.parametrize("userinfo", [USERINFO])
@pytest.mark.parametrize(
    "existing_entries,removed_from",
    [
        ([LDAP_TESTGROUP_ENTRY, LDAP_GROUP1_ENTRY, LDAP_GROUP2_ENTRY], ["testgroup"]),
        (
            [
                LDAP_USER_ENTRY,
                LDAP_TESTGROUP_ENTRY,
                LDAP_GROUP1_ENTRY,
                LDAP_GROUP2_ENTRY,
            ],
            ["testgroup"],
        ),
        (
            [
                LDAP_USER_ENTRY,
                LDAP_PRECREATED_TESTGROUP_ENTRY,
                LDAP_GROUP1_ENTRY,
                LDAP_GROUP2_ENTRY,
            ],
            [],
        ),
    ],
)
@pytest.mark.parametrize("supplementary_group_names", [["group1", "group2"], []])
def test_mod_user_success(ldap_mod_entry, userinfo, removed_from):
    added, removed = ldap_mod_entry["user"].mod(ldap_mod_entry["supplementary_groups"])

    assert set(added) == set(
        [grp.name for grp in ldap_mod_entry["supplementary_groups"]]
    )
    assert set(removed) == set(removed_from)

    for group in ldap_mod_entry["supplementary_groups"]:
        members = group.get_ldap_entry().get("memberUid") or []
        assert userinfo["username"] in members
        assert userinfo["username"] in members


@pytest.mark.parametrize("uid_min,uid_max", [(1000, 2000), (2000, 3000)])
@pytest.mark.parametrize("gid_min,gid_max", [(1000, 2000)])
def test_connection_nextuid(ldap_connection, uid_min):
    assert ldap_connection.get_next_uid() == uid_min
    assert ldap_connection.get_next_uid() == uid_min + 1
    assert ldap_connection.get_next_uid() == uid_min + 2


@pytest.mark.parametrize("uid_min,uid_max", [(1000, 2000)])
@pytest.mark.parametrize("gid_min,gid_max", [(1000, 2000), (2333, 3121)])
def test_connection_nextgid(ldap_connection, gid_min):
    assert ldap_connection.get_next_gid() == gid_min
    assert ldap_connection.get_next_gid() == gid_min + 1
    assert ldap_connection.get_next_gid() == gid_min + 2


@pytest.mark.parametrize("uid_min,uid_max", [(1000, 2000)])
@pytest.mark.parametrize("gid_min,gid_max", [(1000, 2000)])
def test_connection_nextuid_all_taken(ldap_connection, monkeypatch):
    with monkeypatch.context() as mp:
        mp.setattr(ldap_connection, "is_uid_taken", lambda *x: True)
        with pytest.raises(Failure):
            ldap_connection.get_next_uid()


@pytest.mark.parametrize("uid_min,uid_max", [(1000, 2000)])
@pytest.mark.parametrize("gid_min,gid_max", [(1000, 2000)])
def test_connection_nextgid_all_taken(ldap_connection, monkeypatch):
    with monkeypatch.context() as mp:
        mp.setattr(ldap_connection, "is_gid_taken", lambda *x: True)
        with pytest.raises(Failure):
            ldap_connection.get_next_gid()


@pytest.mark.parametrize("uid_min,uid_max", [(1000, 1000)])
@pytest.mark.parametrize("gid_min,gid_max", [(1000, 1000)])
def test_connection_nextuid_out_of_range(ldap_connection):
    ldap_connection.get_next_uid()
    with pytest.raises(Failure):
        ldap_connection.get_next_uid()


@pytest.mark.parametrize("uid_min,uid_max", [(1000, 1000)])
@pytest.mark.parametrize("gid_min,gid_max", [(1000, 1000)])
def test_connection_nextgid_out_of_range(ldap_connection):
    ldap_connection.get_next_gid()
    with pytest.raises(Failure):
        ldap_connection.get_next_gid()


@pytest.mark.parametrize("uid_min,uid_max", [(2000, 1000)])
@pytest.mark.parametrize("gid_min,gid_max", [(2000, 1000)])
def test_connection_nextuid_bad_range(ldap_connection):
    with pytest.raises(Failure):
        ldap_connection.get_next_uid()


@pytest.mark.parametrize("uid_min,uid_max", [(2000, 1000)])
@pytest.mark.parametrize("gid_min,gid_max", [(2000, 1000)])
def test_connection_nextgid_bad_range(ldap_connection):
    with pytest.raises(Failure):
        ldap_connection.get_next_gid()



"""
Manages a user and groups via standard UNIX shadow-utils(8).
"""

# vim: foldmethod=indent : tw=100
# pylint: disable=invalid-name, superfluous-parens
# pylint: disable=logging-fstring-interpolation, logging-not-lazy, logging-format-interpolation
# pylint: disable=raise-missing-from, missing-docstring, too-few-public-methods

import subprocess
from subprocess import CalledProcessError
from pathlib import Path
from os import chown
from datetime import datetime
import logging

# from typing import Union

import regex
from unidecode import unidecode

from ldf_adapter.config import CONFIG
from ldf_adapter.results import Failure
from ldf_adapter.backend import generic

logger = logging.getLogger(__name__)


class User(generic.User):
    def __init__(self, userinfo, **hooks):
        """
        Arguments:
        userinfo -- Only these attributes are used:
                `username` (which is passed through `make_shadow_compatible`),
                `primary_group`
                `unique_id`  stored in gecos, used to find the user
                `ssh_keys`
        """
        super().__init__(userinfo, **hooks)
        self.unique_id = userinfo.unique_id
        logger.debug(f"backend processing: {userinfo.unique_id}")
        if self.exists():
            logger.debug(
                f"This user does actually exist. The name is: {self.get_username()}"
            )
            self.set_username(self.get_username())
        else:
            self.set_username(userinfo.username)

        # should be Group(self.get_primary_group()) when user exists?
        self.primary_group = Group(userinfo.primary_group)
        self.ssh_keys = [key["value"] for key in userinfo.ssh_keys]

    @staticmethod
    def ROOT():
        """ROOT directory for user db
        only used for testing purposes, defaults to '/' otherwise"""
        return "/"

    def exists(self):
        """Check wheter a user (identified by the unique_id) exists"""
        return bool(
            self.unique_id
            in [entry["gecos"] for entry in User.__all_passwd_entries("gecos").values()]
        )

    def is_suspended(self):
        """Optional, only if the backend supports it.
        Inform the user whether a user was suspended (e.g. due to a security incident)
        """
        if self.exists():
            options = ["-l"]
            try:
                result = subprocess.run(
                    ["chage"] + options + [self.name],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    check=True,
                )
            except CalledProcessError as e:
                msg = (e.stderr or e.stdout or b"").decode("utf-8").strip()
                logger.error(
                    "Error executing '{}': {}".format(
                        " ".join(e.cmd), msg or "<no output>"
                    )
                )
                raise Failure(
                    message=f"Cannot get info for user: {msg or '<no output>'}"
                )
            try:
                pattern = regex.compile(r"Account expires\s+: (.*)")
                match = pattern.search(result.stdout.decode("utf-8"))
                if match:
                    expiration_date = match.group(1)
                    if expiration_date == "never":
                        return False
                    expiration_date_sec = int(
                        datetime.strptime(expiration_date, "%b %d, %Y").strftime("%s")
                    )
                    if expiration_date_sec - datetime.now().timestamp() <= 0:
                        return True
            except Exception as e:
                logger.error(e)
                raise Failure(message=f"Could not get: {e or '<no output>'}")
        return False

    def is_limited(self):
        """Optional, only if the backend supports it.
        Inform the user whether a user has limited access"""
        # if shell is nologin
        try:
            shell = self.__passwd_entry.get("shell", None)
            return shell in ["/sbin/nologin", "/usr/bin/nologin", "/bin/nologin"]
        except KeyError:
            return False

    def name_taken(self, name):
        """Check if a username is already taken by *another* user"""
        name = make_shadow_compatible(name)
        taken = name in [
            entry["login"] for entry in User.__all_passwd_entries("login").values()
        ]
        logger.debug(f"name_taken: {taken}")
        return taken and name != self.get_username()

    def get_username(self):
        """Return username based on unique_id"""
        try:
            return User.__all_passwd_entries("gecos")[self.unique_id]["login"]
        except KeyError:
            return None

    def set_username(self, username):
        """Set local username on the service."""
        self.name = make_shadow_compatible(username)

    def get_primary_group(self):
        """Check if a user exists based on unique_id and return the primary group name."""
        return Group.get_group_by_id(self.__gid)

    def _create_cmd(self):
        return [
            "useradd",
            "--comment",
            self.unique_id,
            "-g",
            self.primary_group.name,
            "--shell",
            CONFIG.backend.local_unix.shell,
            "-b",
            CONFIG.backend.local_unix.home_base,
            "-m",
            self.name,
        ]

    def create(self):
        logger.debug(f"Creating user '{self.name}' for {self.unique_id} ")
        try:
            subprocess.run(
                self._create_cmd(),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                check=True,
            )
        except CalledProcessError as e:
            msg = (e.stderr or e.stdout or b"").decode("utf-8").strip()
            logger.error(
                "Error executing '{}': {}".format(" ".join(e.cmd), msg or "<no output>")
            )
            raise Failure(message=f"Cannot create user ({msg or '<no output>'})")

    def create_tostring(self):
        return " ".join(self._create_cmd())

    @staticmethod
    def create_fromstring(create_cmd):
        # some sanity checks
        if not create_cmd.startswith("useradd"):
            raise Failure(
                message=f"Cannot create user (command is not applicable to this backend): {create_cmd}"
            )
        try:
            subprocess.run(
                create_cmd.split(" "),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                check=True,
            )
        except CalledProcessError as e:
            msg = (e.stderr or e.stdout or b"").decode("utf-8").strip()
            logger.error(
                "Error executing '{}': {}".format(" ".join(e.cmd), msg or "<no output>")
            )
            raise Failure(message=f"Cannot create user ({msg or '<no output>'})")

    def update(self):
        pass

    def delete(self):
        if not self.exists():
            raise Failure(
                message=f"Cannot delete user: no user found for {self.unique_id}."
            )

        name = self.__passwd_entry["login"]

        try:
            subprocess.run(
                ["/usr/bin/pkill", "-u", name],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                check=True,
            )
        except CalledProcessError:
            pass
        try:
            subprocess.run(
                ["userdel", name],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                check=True,
            )
        except CalledProcessError as e:
            msg = (e.stderr or e.stdout or b"").decode("utf-8").strip()
            logger.error(
                "Error executing '{}': {}".format(" ".join(e.cmd), msg or "<no output>")
            )
            raise Failure(message=f"Cannot delete user: {msg or '<no output>'}")

    def _mod_cmd(self, supplementary_groups=None):
        """Create command to modify unix user with given groups.

        The user's current groups are merged with the given additional groups,
        and the removal_groups are removed from the list, to generate a list of groups
        that the user should belong to after executing this command.
        A single `usermod` command is created to add and remove the user from all the given groups.

        Args:
            supplementary_groups (list[Group], optional): groups to be set as the user's groups. Defaults to None.

        Returns:
            list: usermod command to modify unix user
        """
        options = []
        if supplementary_groups is not None and supplementary_groups != []:
            logger.debug(
                "Ensuring user '{}' is member of these groups {}".format(
                    self.name, [g.name for g in supplementary_groups]
                )
            )
            options += ["--groups", ",".join([g.name for g in supplementary_groups])]

        return ["usermod"] + options + [self.name]

    def mod(self, supplementary_groups=None):
        """Modify the user on the service.
        After this operation, the user will only be part of the provided groups.

        Arguments:
        supplementary_groups (list[Group], optional): the list of groups the user must be part of. Defaults to None.
            The corresponding unix groups are assumed to exist.

        Returns:
        two lists of groups: the groups the user was added to and the groups the user was removed from
        """
        try:
            group_before = self.get_groups()
            subprocess.run(
                self._mod_cmd(supplementary_groups=supplementary_groups),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                check=True,
            )
            groups_after = self.get_groups()
            groups_added = list(set(groups_after) - set(group_before))
            groups_removed = list(set(group_before) - set(groups_after))
            return groups_added, groups_removed
        except CalledProcessError as e:
            msg = (e.stderr or e.stdout or b"").decode("utf-8").strip()
            logger.error(
                "Error executing '{}': {}".format(" ".join(e.cmd), msg or "<no output>")
            )
            raise Failure(message=f"Cannot modify user: {msg or '<no output>'}")

    def mod_tostring(self, supplementary_groups=None):
        return " ".join(self._mod_cmd(supplementary_groups=supplementary_groups))

    @staticmethod
    def mod_fromstring(mod_cmd):
        # some sanity checks
        if not mod_cmd.startswith("usermod"):
            raise Failure(
                message=f"Cannot modify user (command is not applicable to this backend): {mod_cmd}"
            )
        try:
            subprocess.run(
                mod_cmd.split(" "),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                check=True,
            )
        except CalledProcessError as e:
            msg = (e.stderr or e.stdout or b"").decode("utf-8").strip()
            logger.error(
                "Error executing '{}': {}".format(" ".join(e.cmd), msg or "<no output>")
            )
            raise Failure(message=f"Cannot create user ({msg or '<no output>'})")

    def get_groups(self):
        """Get a list of names of all service groups that the user belongs to.

        If the user doesn't exist, return an empty list.
        """
        try:
            result = subprocess.run(
                ["id", "-Gn", self.name],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                check=True,
            )
            return result.stdout.decode("utf-8").strip().split()
        except CalledProcessError as e:
            msg = (e.stderr or e.stdout or b"").decode("utf-8").strip()
            logger.warning(
                "Error executing '{}': {}".format(" ".join(e.cmd), msg or "<no output>")
            )
            logger.info("Returning empty group list for user %s", self.name)
            return []

    def __expire(self, expiration_date=datetime.today().strftime("%Y-%m-%d")):
        options = ["-E", expiration_date]
        try:
            subprocess.run(
                ["chage"] + options + [self.name],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                check=True,
            )
        except CalledProcessError as e:
            msg = (e.stderr or e.stdout or b"").decode("utf-8").strip()
            logger.error(
                "Error executing '{}': {}".format(" ".join(e.cmd), msg or "<no output>")
            )
            raise Failure(
                message=f"Cannot set expiration date for user: {msg or '<no output>'}"
            )

    def suspend(self):
        self.__expire(datetime.today().strftime("%Y-%m-%d"))

    def resume(self):
        self.__expire("-1")

    def __set_shell(self, shell):
        try:
            subprocess.run(
                ["usermod", "-s", shell, self.name],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                check=True,
            )
        except CalledProcessError as e:
            msg = (e.stderr or e.stdout or b"").decode("utf-8").strip()
            logger.error(
                "Error executing '{}': {}".format(" ".join(e.cmd), msg or "<no output>")
            )
            raise Failure(message=f"Cannot modify user: {msg or '<no output>'}")

    def limit(self):
        self.__set_shell("/sbin/nologin")  # or /usr/sbin/nologin ?

    def unlimit(self):
        self.__set_shell(CONFIG.backend.local_unix.shell)

    def install_ssh_keys(self):
        try:
            if len(self.ssh_keys) > 0:
                if CONFIG.backend.local_unix.deploy_user_ssh_keys:
                    logger.debug(f"Deploying these ssh keys: {self.ssh_keys}")
                    self.__authorized_keys.parent.mkdir(parents=True, exist_ok=True)
                    self.__authorized_keys.parent.chmod(0o700)
                    chown(self.__authorized_keys.parent, self.__uid, self.__gid)

                    self.__authorized_keys.write_text("\n".join(self.ssh_keys))
                    self.__authorized_keys.chmod(0o600)
                    chown(self.__authorized_keys, self.__uid, self.__gid)
        except IOError as e:
            logger.error(e)
            raise Failure(message=f"Could not write new ssh keys: {e or '<no output>'}")
        except Exception as e:
            logger.error(e)
            raise Failure(
                message=f"Cannot change owner or permissions: {e or '<no output>'}"
            )

    def uninstall_ssh_keys(self):
        """Remove any SSH keys stored in the users .authorized_keys file."""
        try:
            self.__authorized_keys.unlink()
        except FileNotFoundError:
            pass

    @property
    def __authorized_keys(self):
        return Path(self.__passwd_entry["home"]) / ".ssh" / "authorized_keys"

    @property
    def __uid(self):
        try:
            return int(self.__passwd_entry.get("uid", None))
        except TypeError as e:
            logger.error(e)
            raise Failure(message=f"Could not get uid: {e or '<no output>'}")

    @property
    def __gid(self):
        try:
            return int(self.__passwd_entry.get("gid", None))
        except TypeError as e:
            logger.error(e)
            raise Failure(message=f"Could not get gid: {e or '<no output>'}")

    @property
    def __passwd_entry(self):
        return User.__all_passwd_entries("gecos").get(self.unique_id, {})

    @staticmethod
    def __all_passwd_entries(ID_FIELD="gecos") -> dict:
        """Get all passwd entries using ID_FIELD to index entries.

        Args:
            ID_FIELD (str, optional): any of the fields defined in PASSWD_FIELDS. Defaults to "name".
        Returns:
            dict: all passwd entries indexed by ID_FIELD, where each entry is represented as a dict with fields PASSWD_FIELDS.
        """
        PASSWD_PATH = Path(User.ROOT()) / "etc" / "passwd"
        PASSWD_FIELDS = ["login", "pw", "uid", "gid", "gecos", "home", "shell"]

        try:
            raw = PASSWD_PATH.read_text()
        except IOError as e:
            logger.error(e)
            raise Failure(
                message="Could not get information about "
                f"existing users on system: {e or '<no output>'}"
            )
        else:
            # for empty file return empty dict
            if raw.strip() == "":
                return {}
            users = [
                dict(zip(PASSWD_FIELDS, line.split(":")))
                for line in raw.strip().split("\n")
            ]

            # import json
            # thedata={user[ID_FIELD]: user for user in users}
            # str_str = json.dumps(thedata, sort_keys=True, indent=4, separators=(',', ': '))
            # logger.debug(str_str)
            # x = {user[ID_FIELD]: user for user in users}
            # logger.debug(F"whattttt: {x}")
            return {user[ID_FIELD]: user for user in users}


class Group(generic.Group):
    def __init__(self, name):
        logger.debug(f"my own group name: {name}")
        if name is None:
            self.name = None
        else:
            self.original_name = name
            if CONFIG.backend.local_unix.punch4nfdi:
                self.name = make_shadow_compatible_punch4nfdi(name)
            else:
                self.name = make_shadow_compatible(name)
            self.name_v004 = make_shadow_compatible_v044(name)

    @staticmethod
    def ROOT():
        """ROOT directory for user db
        only used for testing purposes, defaults to '/' otherwise"""
        return "/"

    def exists(self):
        return bool(self.__group_entry)

    def _create_cmd(self):
        return ["groupadd", self.name]

    def create(self):
        try:
            subprocess.run(
                self._create_cmd(),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                check=True,
            )
        except CalledProcessError as e:
            msg = (e.stderr or e.stdout or b"").decode("utf-8").strip()
            logger.error(
                "Error executing '{}': {}".format(" ".join(e.cmd), msg or "<no output>")
            )
            raise Failure(message=f"Cannot create group: {msg or '<no output>'}")

    def create_tostring(self):
        return " ".join(self._create_cmd())

    @staticmethod
    def create_fromstring(create_cmd):
        # some sanity checks
        if not create_cmd.startswith("groupadd"):
            raise Failure(
                message=f"Cannot create group (command is not applicable to this backend): {create_cmd}"
            )
        try:
            subprocess.run(
                create_cmd.split(" "),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                check=True,
            )
        except CalledProcessError as e:
            msg = (e.stderr or e.stdout or b"").decode("utf-8").strip()
            logger.error(
                "Error executing '{}': {}".format(" ".join(e.cmd), msg or "<no output>")
            )
            raise Failure(message=f"Cannot create group ({msg or '<no output>'})")

    def delete(self):
        # groupdel
        raise NotImplementedError("Do we even need this function?")

    def mod(self):
        # groupmod
        raise NotImplementedError("Do we even need this function?")

    @property
    def members(self):
        members = self.__group_entry.get("members", [])
        logger.debug(f"GROUP MEMBERS: {members}")
        return self.__group_entry.get("members", [])

    @property
    def __group_entry(self):
        return Group.__all_group_entries().get(self.name, {})

    @staticmethod
    def __rename(original_name, old_name, new_name):
        logger.warning(
            f"Local group already exists for {original_name} with an old naming convention."
        )
        logger.warning(f"Renaming group {old_name} to {new_name}.")
        try:
            subprocess.run(["groupmod", "--new-name", new_name, old_name], check=True)
        except CalledProcessError as e:
            msg = (e.stderr or e.stdout or b"").decode("utf-8").strip()
            logger.error(
                "Error executing '{}': {}".format(" ".join(e.cmd), msg or "<no output>")
            )
            raise Failure(message=f"Cannot create group: {msg or '<no output>'}")

    @staticmethod
    def __fix_duplicates(fix_id, original_name, old_name, new_name):
        logger.error(
            f"{fix_id}: Two local groups exist for {original_name}, created with different versions of make_shadow_compatible."
        )
        logger.error(
            f"{fix_id}: Please make sure all files owned by group {old_name} are owned by group {new_name} before removing group {old_name}."
        )

    def fix_group_names(self):
        """Apply fix for groups that were created with different make_shadow_compatible versions.

        3 use cases:
        - group with broken name does not exist: nothing to do
        - group with broken name exists, but new name does not exist yet: rename broken group
        - both group names exist: log this for an admin to merge the groups
        """
        if self.name is None or self.name_v004 is None:
            return
        if self.name == self.name_v004:
            return
        all_groups = Group.__all_group_entries()
        if self.name_v004 in all_groups:
            if self.name not in all_groups:
                Group.__rename(self.original_name, self.name_v004, self.name)
            else:
                Group.__fix_duplicates(
                    "FIX v0.4.4", self.original_name, self.name_v004, self.name
                )

    @staticmethod
    def get_group_by_id(gid):
        """Gets the group name for a given GID."""
        try:
            return Group.__all_group_entries(ID_FIELD="gid")[str(gid)]["name"]
        except KeyError:
            return None

    @staticmethod
    def __all_group_entries(ID_FIELD="name"):
        """Get all group entries using ID_FIELD to index entries.

        Args:
            ID_FIELD (str, optional): any of the fields defined in GROUP_FIELDS. Defaults to "name".
        Returns:
            dict: all group entries indexed by ID_FIELD, where each entry is represented as a dict with fields GROUP_FIELDS.
        """
        GROUP_PATH = Path(Group.ROOT()) / "etc" / "group"
        GROUP_FIELDS = ["name", "password", "gid", "members"]
        LIST_FIELD = "members"

        try:
            raw = GROUP_PATH.read_text()
        except IOError as e:
            logger.error(e)
            raise Failure(
                message="Could not get information about "
                f"existing users on system: {e or '<no output>'}"
            )
        else:
            # for empty file return empty dict
            if raw.strip() == "":
                return {}
            groups = [
                dict(zip(GROUP_FIELDS, line.split(":")))
                for line in raw.strip().split("\n")
            ]

            for group in groups:
                group[LIST_FIELD] = group[LIST_FIELD].split(",")  # type: ignore

            return {group[ID_FIELD]: group for group in groups}


def make_shadow_compatible(orig_word) -> str:
    """Make shadow compatible, using a configured function"""
    mode = CONFIG.backend.local_unix.shadow_compatibility_function
    #  logger.info(f"MODE: {mode}")
    if mode == "v044":
        #  logger.info("v044")
        return make_shadow_compatible_v044(orig_word)
    elif mode == "punch":
        #  logger.info("PUNCH")
        return make_shadow_compatible_punch4nfdi(orig_word)
    elif mode == "default":
        #  logger.info("default")
        return make_shadow_compatible_default(orig_word)


def make_shadow_compatible_default(orig_word) -> str:
    """Ensure that orig_word is a valid user/group name for standard shadow utils.

    While this could in theory be achived by simply substituting all non-allowed chars with a valid
    one, we try to translitare sensibly, so that usernames look nicer and to avoid collisions. See
    inline comments for further details.

    Any change made to the word is logged with level WARNING.

    """
    if orig_word is None:
        ## FIXME: raise ValueError
        return None
        # For some reason "None" still comes in on the docker-compose setup.
        # raise Failure(message="Cannot use username 'None' in make_shadow_compatible")
    # Encode German Umlauts
    word = orig_word.translate(
        str.maketrans(
            {
                "ä": "ae",
                "ö": "oe",
                "ü": "ue",
                "Ä": "Ae",
                "Ö": "Oe",
                "Ü": "Ue",
                "ß": "ss",
                "!": "i",
                "$": "s",
                "*": "x",
                "@": "_at_",
            }
        )
    )

    # Unicode -> Ascii
    word = unidecode(word)

    # Downcase
    word = word.lower()

    # Das ist der doofe part. Für die ganzen Sonderzeichen gibt es nicht wirklich
    # eine transliterierung in [-0-9_a-z], daher nehme ich einfach underscore,
    # was ggf. zu Kollisionen führen kann. Witzig: Shadow erlaubt '$' im namen,
    # aber nur *ganz* am Ende ...
    # word = regex.sub(r'[^-0-9_a-z]', '_', word[:-1]) + regex.sub(r'[^-0-9_a-z$]', '_', word[-1])
    # since we already replace $ with s, no need to check for $ at the end
    word = regex.sub(r"[^-0-9_a-z]", "_", word)

    # Shadow will das Namen mit Kleinbuchstaben oder Underscore anfangen
    if regex.match(r"^[a-z_]", word):
        word = word
    else:
        if len(word) >= 32:
            word = "_" + word[1:]
        else:
            word = "_" + word

    # usernames and group names can only be 32 characters long.
    # My fix is to remove characters a) after the first '_' if there is one.
    #                                b) from the beginning if there is none
    # Also adds two dots as an indicator for where the shortening took place

    excess_chars = len(word) - 32
    if excess_chars > 0:
        if len(word.split("_")) == 1:  # no '_' found:
            word = "__" + word[excess_chars + 2 :]
            logger.warning(f"shortened {orig_word} to {word}")

        elif len(word.split("_")) > 1:  # at least one '_' found:
            fragments = word.split("_")
            if (
                len(fragments[1]) > excess_chars
            ):  # we're fine, we can cut excess chars from fragments alone
                fragments[1] = ".." + fragments[1][excess_chars + 2 :]
                # TODO: fix case when len(fragments[1]) == excess_chars + 1
                word = "_".join(fragments)
                logger.warning(f"Shortended group name: {word} ({len(word)})")
            else:
                logger.error(f"User or group name is too long: {word} ({len(word)})")
                raise (ValueError)
                # TODO: fix case when removing chars from one fragment is not enough
                # to shorten the word
                # i.e. len(fragments[1] <= excess_chars)
            # logger.warning(F"shortened {orig_word} to {word}")

    if word != orig_word:
        logger.debug(
            "Name '{}' changed to '{}' for shadow compatibilty".format(orig_word, word)
        )

    return word


def make_shadow_compatible_punch4nfdi(orig_word) -> str:
    """Ensure that orig_word is a valid user/group name for standard shadow utils.
    Special use case for punch4nfdi.

    Summary of transliteration process:
    - german umlauts are replaced with their phonetic equivalents
    - a few special characters are replaced by sensible equivalents:
        - ! to i
        - $ to s
        - * to x
        - @ to _at_
    - unicode characters are decoded to ascii
    - all other special characters are replaced with _
    - shortening names longer than 32 chars to 32 as follows:
        - fragments (substrings separated by _) are shortened starting with the second fragment,
          then going from the last fragment to the first, until the length 32 is reached
        - the first fragment is shortened by removing characters from the end,
          and adding .. at the end to denote the shortening took place, e.g. "abcdef" -> "abc.."
        - the other fragments are shortened by removing characters from the beginning,
          and adding .. at the beginning to denote the shortening took place, e.g. "abcdef" -> "..def"
        - the length of any fragment has to be > 2 to be considered for shortening
        - names that are still longer than 32 chars after this process will raise a ValueError
    """
    if orig_word is None:
        return None
    # Encode German Umlauts
    word = orig_word.translate(
        str.maketrans(
            {
                "ä": "ae",
                "ö": "oe",
                "ü": "ue",
                "Ä": "Ae",
                "Ö": "Oe",
                "Ü": "Ue",
                "ß": "ss",
                "!": "i",
                "$": "s",
                "*": "x",
                "@": "_at_",
            }
        )
    )

    # Unicode -> Ascii
    word = unidecode(word)

    # Downcase
    word = word.lower()

    # Das ist der doofe part. Für die ganzen Sonderzeichen gibt es nicht wirklich
    # eine transliterierung in [-0-9_a-z], daher nehme ich einfach underscore,
    # was ggf. zu Kollisionen führen kann. Witzig: Shadow erlaubt '$' im namen,
    # aber nur *ganz* am Ende ...
    # word = regex.sub(r'[^-0-9_a-z]', '_', word[:-1]) + regex.sub(r'[^-0-9_a-z$]', '_', word[-1])
    # since we already replace $ with s, no need to check for $ at the end
    word = regex.sub(r"[^-0-9_a-z]", "_", word)

    # Shadow will das Namen mit Kleinbuchstaben oder Underscore anfangen
    if not regex.match(r"^[a-z_]", word):
        word = "_" + word
    if regex.match(r"_-", word):
        word = "_" + word[2:]

    # extract namespace and group from punch4nfdi entitlement
    # require the presence of punch4nfdi in the word to skip username and entitlement not belonging to punch4nfdi
    # error will be triggered for group length exceeding 32 characters
    if regex.search(r"punch4nfdi", word):
        pattern = regex.compile(r"(?P<namespace>\S+)_punch4nfdi(?P<group>\S{1,32}$|\b)")
        word_split = pattern.search(word)
        if word_split:
            word_namespace = word_split.group("namespace")
            word_group = "punch4nfdi" + word_split.group("group")
            logger.warning(
                f"entitlement: {word} - namespace: {word_namespace} - group: {word_group}"
            )
            word = word_group
        else:
            logger.error(
                f"namespace and group could not be extracted from the punch4nfdi entitlement {word} because group has a length exceeding 32 characters"
            )
            raise (ValueError)
    else:
        pattern = regex.compile(r"(?P<word>^\S{1,32}$)")
        word_split = pattern.search(word)
        if not word_split:
            logger.error(
                f"username or entitlement {word} has a length of {len(word)} exceeding 32 characters"
            )
            raise (ValueError)

    return word


def make_shadow_compatible_v044(orig_word) -> str:
    """
    This is the function in feudalAdapter v0.4.4, needed here to rollback the side-effects incurred by it.

    Ensure that orig_word is a valid user/group name for standard shadow utils.

    While this could in theory be achieved by simply substituting all non-allowed chars with a valid
    one, we try to transliterate sensibly, so that usernames look nicer and to avoid collisions. See
    inline comments for further details.

    Summary of transliteration process:
    - german umlauts are replaced with their phonetic equivalents
    - a few special characters are replaced by sensible equivalents:
        - ! to i
        - $ to s
        - * to x
        - @ to _at_
    - unicode characters are decoded to ascii
    - all other special characters are replaced with _
    - shortening names longer than 32 chars to 32 as follows:
        - fragments (substrings separated by _) are shortened starting with the first fragment
          until the length 32 is reached
        - a fragment is shortened by removing the necessary amount of characters from the end,
          and adding .. at the end to denote the shortening took place, e.g. "abcdef" -> "abc.."
        - the length of any fragment has to be > 3 to be considered for shortening
        - the first character of a fragment is always kept, ie. the strongest shortening of "abcdef" will be "a.."
        - names that are still longer than 32 chars after this process will raise a ValueError

    Any change made to the word is logged with level WARNING.

    """
    if orig_word is None:
        return None
        # For some reason "None" still comes in on the docker-compose setup.
        # raise Failure(message="Cannot use username 'None' in make_shadow_compatible")
    # Encode German Umlauts
    word = orig_word.translate(
        str.maketrans(
            {
                "ä": "ae",
                "ö": "oe",
                "ü": "ue",
                "Ä": "Ae",
                "Ö": "Oe",
                "Ü": "Ue",
                "ß": "ss",
                "!": "i",
                "$": "s",
                "*": "x",
                "@": "_at_",
            }
        )
    )

    # Unicode -> Ascii
    word = unidecode(word)

    # Downcase
    word = word.lower()

    # Das ist der doofe part. Für die ganzen Sonderzeichen gibt es nicht wirklich
    # eine transliterierung in [-0-9_a-z], daher nehme ich einfach underscore,
    # was ggf. zu Kollisionen führen kann. Witzig: Shadow erlaubt '$' im namen,
    # aber nur *ganz* am Ende ...
    # word = regex.sub(r'[^-0-9_a-z]', '_', word[:-1]) + regex.sub(r'[^-0-9_a-z$]', '_', word[-1])
    # since we already replace $ with s, no need to check for $ at the end
    word = regex.sub(r"[^-0-9_a-z]", "_", word)

    # Shadow will dass Namen mit Kleinbuchstaben oder Underscore anfangen
    if not regex.match(r"^[a-z_]", word):
        word = "_" + word
    # Marcus does not want to start a username with _-
    if regex.match(r"_-", word):
        word = "_" + word[2:]

    # usernames and group names can only be 32 characters long.
    # split names in fragments and loop over them
    # to progressively remove the characters in excess.
    # .. used to indicate where the shortening took place.
    orig_excess_chars = len(word) - 32
    excess_chars = len(word) - 32
    fragments = word.split("_")

    if excess_chars > 0:
        for n in range(len(fragments)):
            if len(fragments[n]) <= 3:
                continue
            if excess_chars == 0:
                break
            excess_chars += 2

            for nchar in range(len(fragments[n]) - 1):
                fragments[n] = fragments[n][: len(fragments[n]) - 1]
                excess_chars -= 1
                if excess_chars == 0:
                    break

            fragments[n] = fragments[n] + ".."

    if orig_excess_chars > 0:
        if excess_chars > 0:
            logger.error(f"User or group name is too long: {word} ({len(word)})")
            raise (ValueError)
        else:
            if len(fragments) > 1:
                word = "_".join(fragments)
            else:
                word = fragments[0]
    return word



"""
Implement approval workflow for user deployments.
"""

import logging
from typing import List, Optional, Tuple

from ldf_adapter import backend
from ldf_adapter.results import Failure
from ldf_adapter.config import CONFIG
from ldf_adapter.userinfo import UserInfo
from ldf_adapter.approval.models import (
    PendingUser,
    PendingGroup,
    PendingMemberships,
    DeploymentState,
)
from ldf_adapter.approval.db import databases
from ldf_adapter.notifier import notifiers
from ldf_adapter.notifier.generic import NotificationType

logger = logging.getLogger(__name__)


class PendingDeployment:
    """Represents a pending deployment request for a user and its groups.
    Communicates with the pending DB to store and retrieve pending deployments."""

    def __init__(self, userinfo: UserInfo) -> None:
        """Initialise a pending deployment for a federated user.
        If a request already exists for this user, initialise properties.

        Args:
            userinfo (UserInfo): user info of federated user
        """
        self._pending_db = databases.get("sqlite")
        self._notifier = notifiers.get(CONFIG.approval.notifier)
        self.unique_id = userinfo.unique_id
        self._sub = userinfo.sub
        self._iss = userinfo.iss
        self._email = userinfo.email
        self._full_name = userinfo.full_name

        self._user = self._pending_db.get_user(userinfo.unique_id)
        self._memberships = self._pending_db.get_memberships(userinfo.unique_id)
        if self._memberships:
            self._groups = list(
                filter(
                    None,
                    [
                        self._pending_db.get_group(m)
                        for m in self._memberships.supplementary_groups
                    ],
                )
            )
        else:
            self._groups = []

    @property
    def user(self) -> Optional[PendingUser]:
        """Get information about the user pending deployment."""
        return self._user

    @property
    def groups(self) -> List[PendingGroup]:
        """Get information about the groups that need to be created for this pending deployment."""
        return self._groups

    @property
    def memberships(self) -> Optional[PendingMemberships]:
        """Get information about the groups that the user needs to be added to."""
        return self._memberships

    @property
    def sub(self) -> str:
        """Return sub as set in pending db, or in userinfo if user property not set."""
        if self.user:
            return self.user.sub
        return self._sub

    @property
    def iss(self) -> str:
        """Return iss as set in pending db, or in userinfo if user property not set."""
        if self.user:
            return self.user.iss
        return self._iss

    @property
    def email(self) -> Optional[str]:
        """Return email as set in pending db, or in userinfo if user property not set."""
        if self.user:
            return self.user.email
        return self._email

    @property
    def full_name(self) -> Optional[str]:
        """Return full_name as set in pending db, or in userinfo if user property not set."""
        if self.user:
            return self.user.full_name
        return self._full_name

    @property
    def username(self) -> Optional[str]:
        """Return local username as set in pending db, or None if user property not set."""
        if self.user:
            return self.user.username
        return None

    def exists(self) -> bool:
        """Whether there is already an entry in the pending DB for this user."""
        return self.user is not None

    def name_taken(self, name: str) -> bool:
        """Whether the username is already reserved by *another* user."""
        entry = self._pending_db.get_user_by_username(name)
        if (
            entry
            and entry.unique_id != self.unique_id
            and entry.state in [DeploymentState.PENDING, DeploymentState.NOTIFIED]
        ):
            return True
        return False

    def create_user(self, service_user: backend.User) -> bool:  # type: ignore
        """Create pending user entry and add it to db.
        Return False if entry already exists in pending db.
        """
        pending_user = PendingUser(
            unique_id=service_user.unique_id,
            sub=self.sub,
            iss=self.iss,
            email=self.email,
            full_name=self.full_name,
            username=service_user.name,
            state=DeploymentState.PENDING,
            cmd=service_user.create_tostring(),
            infodict={},
        )
        if self._pending_db.add_user(pending_user):
            self._user = pending_user
            return True
        return False

    def update_user(self, service_user: backend.User) -> bool:  # type: ignore
        """Add entry in pending db to update an existing user, only if necessary.
        Return True if there was an update.
        """
        return False

    def create_group(self, group: backend.Group) -> bool:  # type: ignore
        """Create new pending group entry and add it to db.
        Return False if the entry already exists for this user in pending db.
        """
        if group.name in [g.name for g in self.groups]:
            return False
        pending_group = PendingGroup(
            name=group.name,
            state=DeploymentState.PENDING,
            cmd=group.create_tostring(),
            infodict={},
        )
        self._pending_db.add_group(pending_group)
        self._groups.append(pending_group)
        return True

    def mod(
        self, service_user: backend.User, supplementary_groups: List[backend.Group]
    ) -> Tuple[List[str], List[str]]:
        """Create or update info in pending db s.t. the given user is added and removed
        from the given groups.

        If no pending memberships exist, create a new entry for user's pending memberships
        and add it to pending db.
        If pending db contains given user's membership, modify entry only if the lists of groups
        to be added to/removed from have changed.

        Return True if the pending db has been modified and False otherwise.
        """
        supplementary_groups_names = [group.name for group in supplementary_groups]
        current_groups = service_user.get_groups()
        groups_to_add = list(set(supplementary_groups_names) - set(current_groups))
        groups_to_remove = list(set(current_groups) - set(supplementary_groups_names))

        memberships = PendingMemberships(
            unique_id=service_user.unique_id,
            supplementary_groups=supplementary_groups_names,
            removal_groups=[],
            state=DeploymentState.PENDING,
            cmd=service_user.mod_tostring(supplementary_groups=supplementary_groups),
            infodict={},
        )

        if self._memberships is None:
            if groups_to_add != [] or groups_to_remove != []:
                self._pending_db.add_memberships(memberships)
                self._memberships = memberships
            return groups_to_add, groups_to_remove
        elif set(self._memberships.supplementary_groups) != set(
            supplementary_groups_names
        ):
            self._pending_db.update_memberships(memberships)
            self._memberships = memberships
            return groups_to_add, groups_to_remove
        return [], []

    def remove_data(self):
        """Remove from pending db all data associated to this user."""
        if self.user:
            self._pending_db.remove_user(self.user.unique_id)
        for group in self.groups:
            self._pending_db.remove_group(group.name)
        self._pending_db.remove_memberships(self.unique_id)
        self._user = None
        self._groups = []
        self._memberships = None

    def is_pending(self) -> bool:
        """Whether the deployment was requested and is pending approval."""
        return self.user is not None and self.user.state in [
            DeploymentState.PENDING,
            DeploymentState.NOTIFIED,
        ]

    def is_rejected(self) -> bool:
        """Whether the deployment was requested and was rejected."""
        return self.user is not None and self.user.state == DeploymentState.REJECTED

    def mod_pending(self) -> bool:
        """Whether a group membership change is pending approval."""
        return self.memberships is not None and self.memberships.state in [
            DeploymentState.PENDING,
            DeploymentState.NOTIFIED,
        ]

    def accept(self) -> None:
        """Accept this pending deployment and create user and groups (backend-specific)."""
        for group in self.groups:
            backend.Group.create_fromstring(group.cmd)  # type: ignore
        if self.user:
            backend.User.create_fromstring(self.user.cmd)  # type: ignore
        if self.memberships:
            backend.User.mod_fromstring(self.memberships.cmd)  # type: ignore
        self.remove_data()

    def reject(self) -> None:
        """Reject this pending deployment by setting the state of the user in the pending db
        to 'rejected', and remove pending group and membership entries for this user."""
        if self.user:
            self._pending_db.reject_user(self.unique_id)
            self._user = self._pending_db.get_user(self.unique_id)
        for group in self.groups:
            self._pending_db.remove_group(group.name)
        self._groups = []
        if self.memberships:
            self._pending_db.remove_memberships(self.unique_id)
            self._memberships = None

    def _set_notified(self) -> None:
        """Set the state of this pending deployment to 'notified' by setting the state of the user
        and/or its memberships in the pending db to 'notified'.
        """
        if self.user:
            self._pending_db.notify_user(self.unique_id)
        if self.memberships:
            self._pending_db.notify_memberships(self.unique_id)
        for group in self.groups:
            self._pending_db.notify_group(group.name)

    def notify(self):
        """Send out notification to the service admin and the user about this pending deployment."""
        admin_notification = NotificationType.NONE
        user_notification = NotificationType.NONE
        if self.user and self.user.state == DeploymentState.PENDING:
            # new deployment request
            admin_notification = NotificationType.ADMIN_DEPLOY
            user_notification = NotificationType.USER_DEPLOY
        elif self.memberships and self.memberships.state == DeploymentState.PENDING:
            # only membership changes are pending
            if self.user is None:
                # user exists, but groups need to be updated
                admin_notification = NotificationType.ADMIN_UPDATE
                user_notification = NotificationType.USER_UPDATE
            elif self.user and self.user.state == DeploymentState.NOTIFIED:
                # deployment request already notified, but groups have changed and need to be updated
                admin_notification = NotificationType.ADMIN_DEPLOY_UPDATE
                user_notification = NotificationType.ADMIN_DEPLOY_UPDATE
        # build notification data
        user_cmd = self.user.cmd if self.user else ""
        groups_cmd = "\n".join([m.cmd for m in self.groups])
        memberships_cmd = self.memberships.cmd if self.memberships else ""
        data = {
            "hostname": CONFIG.login_info.ssh_host,
            "unique_id": self.unique_id,
            "full_name": self.full_name,
            "email": self.email,
            "user_cmd": user_cmd,
            "groups_cmd": groups_cmd,
            "memberships_cmd": memberships_cmd,
            "sub": self.sub,
            "iss": self.iss,
            "username": self.username,
        }
        try:
            notified = self._notifier.notify(
                notification_type=admin_notification, data=data
            )
            if self.email:
                self._notifier.notify(notification_type=user_notification, data=data)
            else:
                logger.warning(
                    "No email found in deployment request: user will not receive email notification."
                )
            if notified:
                self._set_notified()
        except Exception as ex:
            logger.error(f"Failed to send notification: {ex}")
            raise Failure(
                message="Failed to notify admin of deployment request. Please contact an admin or try again later."
            )

    def test_notifier(self):
        self._notifier.test()
